import time
from typing import Union, Dict, List, Optional
from logging import Logger
from stpstone.utils.connections.netops.proxies._free.proxy_nova import ProxyNova
from stpstone.utils.connections.netops.proxies._free.proxy_scrape import ProxyScrapeAll, ProxyScrapeCountry
from stpstone.utils.connections.netops.proxies._free.proxy_webshare import ProxyWebShare
from stpstone.utils.connections.netops.proxies._free.freeproxy_world import FreeProxyWorld
from stpstone.utils.connections.netops.proxies._free.free_proxy_list_net import FreeProxyNet
from stpstone.utils.connections.netops.proxies._free.spysme import SpysMeCountries
from stpstone.utils.connections.netops.proxies._free.spysone import SpysOneCountry
from stpstone.utils.loggs.create_logs import CreateLog


class YieldFreeProxy:

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None,
        str_plan_id_webshare: str = "free",
        max_iter_find_healthy_proxy: int = 10,
        timeout_session: Optional[float] = 1000.0,
        int_wait_load_seconds: Optional[int] = 10,
    ) -> None:
        self.bl_new_proxy = bl_new_proxy
        self.dict_proxies = dict_proxies
        self.int_retries = int_retries
        self.int_backoff_factor = int_backoff_factor
        self.bl_alive = bl_alive
        self.list_anonymity_value = list_anonymity_value
        self.list_protocol = list_protocol
        self.str_continent_code = str_continent_code
        self.str_country_code = str_country_code
        self.bl_ssl = bl_ssl
        self.float_min_ratio_times_alive_dead = float_min_ratio_times_alive_dead
        self.float_max_timeout = float_max_timeout
        self.bl_use_timer = bl_use_timer
        self.list_status_forcelist = list_status_forcelist
        self.logger = logger
        self.str_plan_id_webshare = str_plan_id_webshare
        self.max_iter_find_healthy_proxy = max_iter_find_healthy_proxy
        self.timeout_session = timeout_session
        self.int_wait_load_seconds = int_wait_load_seconds
        self.create_logs = CreateLog()

        self.cls_spys_one_country = SpysOneCountry(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_free_proxy_net = FreeProxyNet(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_spysme_countries = SpysMeCountries(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_proxy_nova = ProxyNova(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_proxy_scrape_all = ProxyScrapeAll(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_proxy_scrape_country = ProxyScrapeCountry(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_proxy_webshare = ProxyWebShare(
            str_plan_id=str_plan_id_webshare,
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )

        self.cls_freeproxy_world = FreeProxyWorld(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger,
            int_wait_load_seconds=int_wait_load_seconds
        )

        self._retry_count = 0
        self.cached_sessions = self._cache
        self._last_cache_time = time.time()

    @property
    def _cache(self) -> List[Dict[str, str]]:
        list_ser = list()
        for list_ in [
            self.cls_spys_one_country.configured_sessions,
            self.cls_free_proxy_net.configured_sessions,
            self.cls_spysme_countries.configured_sessions,
            self.cls_freeproxy_world.configured_sessions,
            self.cls_proxy_nova.configured_sessions,
            self.cls_proxy_scrape_all.configured_sessions,
            self.cls_proxy_scrape_country.configured_sessions,
            self.cls_proxy_webshare.configured_sessions,
        ]:
            if list_ is not None:
                list_ser.extend(list_)
        self.create_logs.log_message(
            self.logger,
            f"Number of proxies healthy: {len(list_ser)}",
            log_level="info"
        )
        if (len(list_ser) == 0) and (self._retry_count < self.max_iter_find_healthy_proxy):
            self.create_logs.log_message(
                self.logger,
                f"*** No proxies available - retrying {self._retry_count+1}/{self.max_iter_find_healthy_proxy}",
                log_level="warning"
            )
            self._retry_count += 1
        if (len(list_ser) == 0) and (self._retry_count >= self.max_iter_find_healthy_proxy):
            self.create_logs.log_message(
                self.logger,
                f"*** No proxies available after {self.max_iter_find_healthy_proxy} attempts",
                log_level="error"
            )
            raise ValueError("No proxies available")
        if len(list_ser) > 0:
            self._last_cache_time = time.time()
            self._retry_count = 0
        return list_ser

    def __next__(self) -> Dict[str, str]:
        if self.bl_new_proxy == False: return None
        list_proxies = list()
        while (len(list_proxies) == 0) \
            and (time.time() - self._last_cache_time < self.timeout_session):
            list_proxies = self.cached_sessions
            if len(list_proxies) == 0: self.cached_sessions = self._cache
        if (len(list_proxies) == 0) \
            and (time.time() - self._last_cache_time > self.timeout_session):
            self.create_logs.log_message(
                self.logger,
                f"*** Timeout reached: {self.timeout_session} seconds / Retries: {self._retry_count}",
                log_level="critical"
            )
            raise ValueError("No proxies available")
        if len(list_proxies) == 0:
            self.create_logs.log_message(
                self.logger,
                f"*** No proxies available after {self.max_iter_find_healthy_proxy} attempts",
                log_level="critical"
            )
            raise ValueError("No proxies available")
        return list_proxies.pop(0)



### CONNECTING TO SQLITE DATABASE ###

import sqlite3
import backoff
from logging import Logger
import pandas as pd
from typing import List, Dict, Any, Optional
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.parsers.json import JsonFiles
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.parsers.pd import DealingPd
from stpstone.utils.connections.databases.abc import DatabaseConnection


class SQLiteDB(DatabaseConnection):

    def __init__(self, db_path:str, logger:Optional[Logger]=None) -> None:
        """
        DOCSTRING: INITIALIZES THE CONNECTION TO THE SQLITE DATABASE
        INPUTS: DB_PATH
        OUTPUTS: -
        """
        self.db_path = db_path
        self.logger = logger
        self.conn:sqlite3.Connection = sqlite3.connect(self.db_path)
        self.cursor:sqlite3.Cursor = self.conn.cursor()

    @backoff.on_exception(
        backoff.constant,
        sqlite3.OperationalError,
        interval=10,
        max_tries=20,
    )
    def execute(self, str_query:str) -> None:
        """
        DOCSTRING: RUN QUERY WITH DML ACCESS
        INPUTS: QUERY
        OUTPUTS: -
        """
        self.cursor.execute(str_query)

    @backoff.on_exception(
        backoff.constant,
        sqlite3.OperationalError,
        interval=10,
        max_tries=20,
    )
    def read(self, str_query:str, dict_dtypes:Dict[str, Any],
        list_cols_dt:Optional[List[str]]=None) -> pd.DataFrame:
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # retrieving dataframe
        df_ = pd.read_sql_query(str_query, self.conn)
        if df_.empty == False:
            #   changing data types
            df_ = DealingPd().pipeline_df_startup(df_, dict_dtypes, list_cols_dt)
        # return dataframe
        return df_

    @backoff.on_exception(
        backoff.constant,
        sqlite3.OperationalError,
        interval=10,
        max_tries=20,
    )
    def insert(self, json_data:List[Dict[str, Any]], str_table_name:str,
        bl_insert_or_ignore:bool=False) -> None:
        """
        DOCSTRING: INSERTS DATA FROM A JSON OBJECT INTO A SQLITE TABLE
        INPUTS: JSON_DATA
        OUTPUTS: -
        """
        # validate json, in order to have the same keys
        json_data = JsonFiles().normalize_json_keys(json_data)
        # sql insert statement
        list_columns = ', '.join(json_data[0].keys())
        list_data = ', '.join(['?' for _ in json_data[0]])
        if bl_insert_or_ignore == True:
            str_query = f'INSERT OR IGNORE INTO {str_table_name} ' \
                + f'({list_columns}) VALUES ({list_data})'
        else:
            str_query = f'INSERT INTO {str_table_name} ' \
                + f'({list_columns}) VALUES ({list_data})'
        try:
            # insert each record
            for record in json_data:
                self.cursor.execute(str_query, tuple(record.values()))
            self.conn.commit()
            if self.logger is not None:
                CreateLog().info(
                    self.logger,
                    f'Succesful commit in db {self.db_path} '
                    + f'/ table {str_table_name}.'
                )
        except Exception as e:
            self.conn.rollback()
            self.close
            if self.logger is not None:
                CreateLog().error(
                    self.logger,
                    'ERROR WHILE INSERTING DATA\n'
                    + f'DB_PATH: {self.db_path}\n'
                    + f'TABLE_NAME: {str_table_name}\n'
                    + f'JSON_DATA: {json_data}\n'
                    + f'ERROR_MESSAGE: {e}'
                )
            raise Exception(
                'ERROR WHILE INSERTING DATA\n'
                + f'DB_PATH: {self.db_path}\n'
                + f'TABLE_NAME: {str_table_name}\n'
                + f'JSON_DATA: {json_data}\n'
                + f'ERROR_MESSAGE: {e}'
            )

    @property
    def close(self) -> None:
        """
        DOCSTRING: CLOSES THE CONNECTION TO THE DATABASE
        INPUTS: -
        OUTPUTS: -
        """
        self.conn.close()



import re
import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Union
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_ANBIMA_INFOS
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.str import StrHandler


class AnbimaInfos(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_ANBIMA_INFOS,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.list_slugs = list_slugs
        self.dt_ref_yymmdd = dt_ref.strftime('%y%m%d')

    def list_rows_injection(self, resp_req: Response) -> Union[List[Any], List[Any]]:
        list_rows_1 = []
        list_rows_2 = []
        raw_text = re.sub(r'--(?=\d)', '--@', resp_req.text)
        raw_text = raw_text.replace(",", ".").replace("--", "-99999").replace("\r", "")
        list_rows_raw = raw_text.split("\n")
        for i, row in enumerate(list_rows_raw):
            if row.startswith("0@") or row.startswith("1@TOTAIS") or row.startswith("2@COMPOSI") \
                or row.startswith("1@Data de Refer") or row.startswith("2@Data de Refer") \
                or len(row) <= 1:
                continue
            elif (row.startswith("1@")) \
                and (not row.startswith("1@TOTAIS")) \
                and (not row.startswith("1@Data de Refer")):
                list_rows_1.append(row[2:])
            elif (row.startswith("2@")) \
                and (not row.startswith("2@COMPOSI")) \
                and (not row.startswith("2@Data de Refer")):
                list_rows_2.append(row[2:])
            elif any(row.startswith(f"{n}@") for n in range(3, 10)):
                raise ValueError(f"ROW #{i}: Unexpected row prefix found: {row}")
            elif "@" not in row:
                raise ValueError(f"ROW #{i}: Unexpected row format found: {row}")
            else:
                raise ValueError(f"ROW #{i}: Unexpected row format found: {row}")
        return list_rows_1, list_rows_2

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        if StrHandler().match_string_like(resp_req.url, '*#source=ima_p2_pvs*'):
            source = "ima_p2_pvs"
            int_idx_row = 1
        elif StrHandler().match_string_like(resp_req.url, '*#source=ima_p2_th_portf*'):
            source = "ima_p2_th_portf"
            int_idx_row = 2
        else:
            source = None
            int_idx_row = None
        if source is None or int_idx_row is None:
            return None
        else:
            list_rows_1, list_rows_2 = self.list_rows_injection(resp_req)
            data_rows = locals()[f"list_rows_{int_idx_row}"]
            df_ = pd.DataFrame([row.split("@") for row in data_rows],
                            columns=list(YAML_ANBIMA_INFOS[source]["dtypes"].keys()))
        return df_



### API TO GRANT ACCESS TO B3 MARGIN SIMULATOR ###

# pypi.org libs
import numpy as np
from requests import request
from typing import List, Dict, Union
# local libs
from stpstone.utils.parsers.json import JsonFiles


class MarginSimulatorB3:

    def __init__(self, dict_payload:List[Dict[str, Union[str, int, float]]],
                 token:str='79a4413f55d7d982b61c669e6dd35eea',
                 host:str='https://simulador.b3.com.br/api/cors-app'):
        """
        INPUTS: DATA PAYLOAD PORTFOLIO EXEMPLE - dict_payload = [
                {
                    'Security': {'symbol': 'ABCBF160'}
                    'SecurityGroup': {'positionTypeCode': 0},
                    'Position': {'longQuantity':100,'shortQuantity': 0,'longPrice': 0,'shortPrice': 0}
                },
                {
                    'Security': {'symbol': 'ABCBF179'},
                    'SecurityGroup': {'positionTypeCode': 0},
                    'Position': {'longQuantity': 100,'shortQuantity': 0,'longPrice': 0,'shortPrice': 0}
                },
                {
                    'Security': {'symbol': 'ABCBF182'},
                    'SecurityGroup': {'positionTypeCode': 0},
                    'Position': {'longQuantity': 0,'shortQuantity':200,'longPrice': 0,'shortPrice': 0}
                }
            ]
        """
        self.dict_payload = dict_payload
        self.token = token,
        self.host = host

    @property
    def total_deficit_surplus(self, method:str='POST', app:str='/web/V1.0/RiskCalculation',
                              value_liquidity_resource:np.int64=4_700_000_000, bl_verify:bool=False,
                              bl_parse_dict_payload_data:bool=True) \
                                -> List[Dict[str, Union[str, int, float]]]:
        """
        REFERENCES: https://simulador.b3.com.br/
        DOCSTRING: TOTAL DEFICIT SURPLUS B3 MARGIN CALL CALCULATION
        INPUTS: METHOD (DEFAULT), KEY HEADER (DEFAULT), URL AUTHENTIFICATION
            HEADER (DEFAULT)
        OUTPUTS: STRING
        """
        # requesting authorization authheader
        dict_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        # payload
        dict_payload = {
            'ReferenceData': {'referenceDataToken': '{}'.format(self.token)},
            'LiquidityResource': {'value': value_liquidity_resource},
            'RiskPositionList': self.dict_payload
        }
        # check wheter it is needed to parse the params dictionary
        if bl_parse_dict_payload_data == True:
            if dict_payload != None:
                dict_payload = JsonFiles().dict_to_json(dict_payload)
        # resquet host REST information
        resp_req = request(method=method, host=self.host + app,
                           headers=dict_headers, data=dict_payload, verify=bl_verify)
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # getting authheader
        return resp_req.json()



### CVM DATA - https://dados.cvm.gov.br/dados

import pandas as pd
import numpy as np
from io import StringIO
from typing import Dict, Any
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.parsers.folders import DirFilesManagement


class CVMDATA:

    def __init__(
        self,
        str_host_cvm:str='https://dados.cvm.gov.br/dados/',
        logger:object=None,
        str_dt_error:str='2100-01-01',
        str_format_dt_input:str='YYYY-MM-DD',
        int_val_err:int=0,
        dict_fund_class_subclass_register:Dict[str, Any]=None
    ):
        self.str_host_cvm = str_host_cvm
        self.logger = logger
        self.str_dt_error = str_dt_error
        self.str_format_dt_input = str_format_dt_input
        self.int_val_err = int_val_err
        self.dict_fund_class_subclass_register = self.funds_classes_subclasses_register_raw \
            if dict_fund_class_subclass_register is None else dict_fund_class_subclass_register

    @property
    def funds_register(
        self,
        str_app:str='FI/CAD/DADOS/cad_fi.csv'
    ):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # url
        url = f'{self.str_host_cvm}{str_app}'
        # read the csv file from the url into a pandas dataframe
        reader = pd.read_csv(url, sep=';', encoding='latin1', decimal='.', thousands=',')
        df_funds_register = pd.DataFrame(reader)
        # validate the content of dataframe
        if (df_funds_register.empty == True) \
            and (self.logger is not None):
            CreateLog().error(
                self.logger, 'Error reading funds register within url: {}'.format(url))
            raise Exception('Error reading funds register within url: {}'.format(url))
        elif df_funds_register.empty == True:
            raise Exception('Error reading funds register within url: {}'.format(url))
        # fill na values
        list_cols = list(df_funds_register.columns)
        list_cols_dts = [
            'DT_REG',
            'DT_CONST',
            'DT_CANCEL',
            'DT_INI_SIT',
            'DT_INI_ATIV',
            'DT_INI_EXERC',
            'DT_FIM_EXERC',
            'DT_INI_CLASSE',
            'DT_PATRIM_LIQ'
        ]
        for col_dt in list_cols_dts:
            df_funds_register[col_dt] = df_funds_register[col_dt].fillna(self.str_dt_error)
        for col_ in [c for c in list_cols if c not in list_cols_dts]:
            df_funds_register[col_] = df_funds_register[col_].fillna(self.int_val_err)
        # changing coumn datatypes
        df_funds_register = df_funds_register.astype({
            'TP_FUNDO': 'category',
            'CNPJ_FUNDO': str,
            'DENOM_SOCIAL': str,
            'CD_CVM': np.int64,
            'DT_REG': str,
            'DT_CONST': str,
            'CD_CVM': int,
            'DT_CANCEL': str,
            'SIT': str,
            'DT_INI_SIT': str,
            'DT_INI_ATIV': str,
            'DT_INI_EXERC': str,
            'DT_FIM_EXERC': str,
            'CLASSE': str,
            'DT_INI_CLASSE': str,
            'RENTAB_FUNDO': str,
            'CONDOM': str,
            'FUNDO_COTAS': str,
            'FUNDO_EXCLUSIVO': str,
            'TRIB_LPRAZO': str,
            'PUBLICO_ALVO' : str,
            'ENTID_INVEST': str,
            'TAXA_PERFM': str,
            'INF_TAXA_PERFM': str,
            'TAXA_ADM': str,
            'INF_TAXA_ADM': str,
            'VL_PATRIM_LIQ': float,
            'DT_PATRIM_LIQ': str,
            'DIRETOR': str,
            'CNPJ_ADMIN': str,
            'ADMIN': str,
            'PF_PJ_GESTOR': str,
            'CPF_CNPJ_GESTOR': str,
            'CNPJ_AUDITOR': str,
            'AUDITOR': str,
            'CNPJ_CUSTODIANTE': str,
            'CUSTODIANTE': str,
            'CNPJ_CONTROLADOR': str,
            'CONTROLADOR': str,
            'INVEST_CEMPR_EXTER': str,
            'CLASSE_ANBIMA': str
        })
        for col_dt in list_cols_dts:
            df_funds_register[col_dt] = [
                DatesBR().str_date_to_datetime(d, self.str_format_dt_input)
                for d in df_funds_register[col_dt]
            ]
        # return the dataframe
        return df_funds_register

    @property
    def funds_classes_subclasses_register_raw(
        self,
        str_app:str='FI/CAD/DADOS/registro_fundo_classe.zip'
    ):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # setting variables
        dict_ = dict()
        # url
        url = f'{self.str_host_cvm}{str_app}'
        # downloading zip file into memory
        list_main_zip = RemoteFiles().get_zip_from_web_in_memory(
            url,
            bl_io_interpreting=False,
            bl_verify=False
        )
        # iterate through files in the ZIP
        for file_info in list_main_zip:
            #   extracting csv files
            if file_info.name.endswith('.csv'):
                dict_[file_info.name] = file_info.read()
        return dict_

    def funds_raw_infos(
        self,
        key_file_name,
        list_cols_dts,
        dict_cols_types
    ):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # assuming self.dict_fund_class_subclass_register[key_file_name] is a bytes object
        file_data = self.dict_fund_class_subclass_register[key_file_name]
        # trying to decode with 'ISO-8859-1' or 'latin1' encoding
        try:
            file_data_str = StringIO(file_data.decode('utf-8'))
        except UnicodeDecodeError:
            file_data_str = StringIO(file_data.decode('ISO-8859-1'))
        # reading csv
        df_ = pd.read_csv(file_data_str, delimiter=';')
        # fill na values
        list_cols = list(df_.columns)
        for col_dt in list_cols_dts:
            df_[col_dt] = df_[col_dt].fillna(self.str_dt_error)
        for col_ in [c for c in list_cols if c not in list_cols_dts]:
            df_[col_] = df_[col_].fillna(self.int_val_err)
        # changing datatypes
        df_ = df_.astype(dict_cols_types)
        for col_dt in list_cols_dts:
            df_[col_dt] = [
                DatesBR().str_date_to_datetime(d, self.str_format_dt_input)
                for d in df_[col_dt]
            ]
        # return the dataframe
        return df_

    @property
    def funds_classes(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # getting dataframe with funds
        df_funds_classes = self.funds_raw_infos(
            'registro_classe.csv',
            [
                'Data_Registro',
                'Data_Constituicao',
                'Data_Inicio',
            ],
            {
                'ID_Registro_Fundo': np.int64,
                'ID_Registro_Classe': np.int64,
                'CNPJ_Classe': str,
                'Codigo_CVM': np.int64,
                'Data_Registro': str,
                'Data_Constituicao': str,
                'Data_Inicio': str,
                'Tipo_Classe': str,
                'Denominacao_Social': str,
                'Situacao': str,
                'Classificacao': str,
                'Indicador_Desempenho': str,
                'Classe_Cotas': str,
                'Classificacao_Anbima': str,
                'Tributacao_Longo_Prazo': str,
                'Entidade_Investimento': str,
                'Permitido_Aplicacao_CemPorCento_Exterior': str,
                'Classe_ESG': str,
                'Forma_Condominio': str,
                'Exclusivo': str,
                'Publico_Alvo': str,
                'CNPJ_Auditor': str,
                'Auditor': str,
                'CNPJ_Custodiante': str,
                'Custodiante': str,
                'CNPJ_Controlador': str,
                'Controlador': str
            }
        )
        # return the dataframe
        return df_funds_classes

    @property
    def funds_register_2(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        df_funds_register = self.funds_raw_infos(
            'registro_fundo.csv',
            [
                'Data_Registro',
                'Data_Constituicao',
                'Data_Cancelamento',
                'Data_Inicio_Situacao',
                'Data_Adaptacao_RCVM175',
                'Data_Inicio_Exercicio_Social',
                'Data_Fim_Exercicio_Social',
                'Data_Patrimonio_Liquido'
            ],
            {
                'ID_Registro_Fundo': np.int64,
                'CNPJ_Fundo': str,
                'Codigo_CVM': np.int64,
                'Data_Registro': str,
                'Data_Constituicao': str,
                'Tipo_Fundo': str,
                'Denominacao_Social': str,
                'Data_Cancelamento': str,
                'Situacao': str,
                'Data_Inicio_Situacao': str,
                'Data_Adaptacao_RCVM175': str,
                'Data_Inicio_Exercicio_Social': str,
                'Data_Fim_Exercicio_Social': str,
                'Patrimonio_Liquido': np.float64,
                'Data_Patrimonio_Liquido': str,
                'Diretor': str,
                'CNPJ_Administrador': str,
                'Administrador': str,
                'Tipo_Pessoa_Gestor': str,
                'CPF_CNPJ_Gestor': str,
                'Gestor': str
            }
        )
        return df_funds_register

    @property
    def funds_subclasses(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        df_funds_subclasses = self.funds_raw_infos(
            'registro_subclasse.csv',
            [
                'Data_Constituicao',
                'Data_Inicio'
            ],
            {
                'ID_Registro_Classe': np.int64,
                'ID_Subclasse': str,
                'Codigo_CVM': pd.Int64Dtype(),
                'Data_Constituicao': str,
                'Data_Inicio': str,
                'Denominacao_Social': str,
                'Situacao': str,
                'Forma_Condominio': str,
                'Exclusivo': str,
                'Publico_Alvo': str
            }
        )
        return df_funds_subclasses



### ASYNCHRONOUS PIPELINES ###

import asyncio
from typing import Callable, List, Any


async def asyncpipeline(data:Any, functions:List[Callable]) -> Any:
    """
    Executes a sequence of asynchronous functions on a given input
    Examples of usage:
        API requests, web scraping, database queries
    Args:
        data (Any): The initial input data
        functions (List[Callable]): A list of async functions
    Returns:
        Any: The processed data.
    """
    for func in functions:
        try:
            data = await func(data)
        except Exception as e:
            print(f"Error in {func.__name__}: {e}")
            break
    return data


if __name__ == "__main__":

    async def async_step_1(data):
        await asyncio.sleep(1)
        return data * 2

    async def async_step_2(data):
        await asyncio.sleep(1)
        return data + 10

    async def main():
        result = await asyncpipeline(5, [async_step_1, async_step_2])
        print(result)  # Output: 20

    asyncio.run(main())



import pandas as pd
from datetime import datetime
from typing import Optional
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from stpstone._config.global_slots import YAML_WW_CRYPTO_COINMARKET
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests


class CoinMarket(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None,
        token:Optional[str]=None
    ) -> None:
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        super().__init__(
            dict_metadata=YAML_WW_CRYPTO_COINMARKET,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token
        )

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        list_ser = list()
        json_ = resp_req.json()
        for dict_ in json_['data']:
            list_ser.append({
                'ID': dict_['id'],
                'NAME': dict_['name'],
                'SYMBOL': dict_['symbol'],
                'PRICE': dict_['quote']['USD']['price'],
                'MARKET_CAP': dict_['quote']['USD']['market_cap'],
                'VOLUME': dict_['quote']['USD']['volume_24h'],
                'SLUG': dict_['slug'],
                'TOTAL_SUPPLY': dict_['total_supply'],
                'CMC_RANK': dict_['cmc_rank'],
                'NUM_MARKET_PAIRS': dict_['num_market_pairs'],
                'LAST_UPDATE': dict_['last_updated'],
            })
        return pd.DataFrame(list_ser)



### CALCULADORA DE TESOURO DIRETO ###

from datetime import date, datetime
import sys
sys.path.append(r'C:\Users\Guilherme\OneDrive\Dev\Python\Packages')
from stpstone.finance.performance_apprraisal.financial_math import FinancialMath
from stpstone.utils.cals.handling_dates import DatesBR


class PrecificacaoTD:

    def ltn(self, ytm, du, valor_nominal=1000, du_1_ano=252, periodicidade_capitalizacao=1):
        """
        DOCSTRING: TÍTULO PRÉ-FIXADO COM PAGAMENTO DE PRINCIPAL E JUROS NO VENCIMENTO
        INPUTS: YIELD TO MATURITY, OU MARCAÇÃO A MERCADO, DU ATÉ O VENCIMENTO, VALOR NOMINAL,
            POR PADRÃO R$ 1.000,00 E DU
        OUTPUTS: VALOR PRESENTE DO TÍTULO, OU MARCAÇÃO A MERADO DO MESMO
        """
        return FinancialMath().present_value(
            FinancialMath().compound_interest(
                ytm, du_1_ano, periodicidade_capitalizacao), du, 0,
            valor_nominal)

    def ntn_f(self, ytm, lista_dus_fluxos_caixa, taxa_nominal_cupom=0.1, valor_nominal=1000,
              du_1_ano=252, periodicidade_pagmento_cupons=126, periodicidade_capitalizacao=1):
        """
        DOCSTRING: TÍTULO PRÉ-FIXADO COM PAGAMENTO DE JUROS SEMESTRALMENTE E PRINCIPAL NO
            VENCIMENTO
        INPUTS: YIELD DO MATURITY, DU ATÉ O VENCIMENTO, CUPOM PAGO AO ANO
            (PERCENTUAL DO VALOR NOMINAL),
        OUTPUTS: VALOR PRESENTE DO FLUXO DE CAIXA
        """
        ytm_real = FinancialMath().compound_interest(
            ytm, du_1_ano, periodicidade_capitalizacao)
        cupom_semestral = valor_nominal \
            * FinancialMath().compound_interest(taxa_nominal_cupom, du_1_ano,
                                                periodicidade_pagmento_cupons)
        print(cupom_semestral)
        print(lista_dus_fluxos_caixa)
        list_fluxos_caixa = [FinancialMath().present_value(ytm_real, du, 0, cupom_semestral)
                             for du in lista_dus_fluxos_caixa]
        list_fluxos_caixa.append(FinancialMath().present_value(ytm_real, lista_dus_fluxos_caixa[-1],
                                                               0, valor_nominal))
        # pu
        return {
            'lista_dus_ntnf': lista_dus_fluxos_caixa,
            'lista_fluxos_caixa_ntnf': list_fluxos_caixa,
            'retorno_bruto_ntnf': sum(list_fluxos_caixa)
        }

    def pr1(self, data_negociacao=DatesBR().add_working_days(DatesBR().curr_date(),
                                                             1).strftime('%d/%m/%Y'),
            dia_atualizacao_vna=15):
        """
        DOCSTRING: PROPORÇÃO ENTRE OS DIAS CORRIDOS DESDE O ÚLTIMO IPCA ATÉ A DATA DE VIGÊNCIA DO
            PRÓXIMO
        INPUTS: DATA DE NEGOCIAÇÃO EM DD/MM/AAAA
        OUTPUTS: INT
        """
        # convertendo a data str para formato datetime
        data_negociacao = DatesBR().str_date_to_datetime(data_negociacao, 'DD/MM/AAAA')
        data_negociacao = DatesBR().add_working_days(data_negociacao, 1)
        # data de negociação antes da data de divulgação do vna, mo mês corrente
        if date(data_negociacao.year, data_negociacao.month,
                dia_atualizacao_vna) > data_negociacao:
            data_proximo_vna = date(
                data_negociacao.year, data_negociacao.month, dia_atualizacao_vna)
            if data_negociacao.month == 1:
                data_vna_anterior = date(
                    data_negociacao.year - 1, 12, dia_atualizacao_vna)
            else:
                data_vna_anterior = date(
                    data_negociacao.year, data_negociacao.month - 1, dia_atualizacao_vna)
        # data de negociação depois da data de divulgação do vna, mo mês corrente
        else:
            if data_negociacao.month == 12:
                data_proximo_vna = date(
                    data_negociacao.year + 1, 1, dia_atualizacao_vna)
            else:
                data_proximo_vna = date(
                    data_negociacao.year, data_negociacao.month + 1, dia_atualizacao_vna)
            data_vna_anterior = date(
                data_negociacao.year, data_negociacao.month, dia_atualizacao_vna)
        # retornar pr1
        return DatesBR().delta_calendar_days(data_negociacao, data_vna_anterior) \
            / DatesBR().delta_calendar_days(data_proximo_vna, data_vna_anterior)

    def vna_projetado_ntnb(self, vna_ultimo_disponivel_ntnb, ipca_projetado_aa, pr1,
                           nper_dias_uteis_aa=252, nper_dias_corridos_am=30):
        """
        DOCSTRING: VNA PROJETADO PARA DIAS DE NEGOCIAÇÃO QUE NÃO TENHA O VNA DISPONÍVEL
        INPUTS: VNA ÚLTIMO DISPONÍVEL (TRUNCADO NA SEXTA CASA), IPCA PROJETADO E PR1
        OUTPUTS: VNA PROJETADO (BOOLEAN)
        """
        ipca_projetado_adu = FinancialMath().compound_interest(ipca_projetado_aa, nper_dias_uteis_aa,
                                                               nper_dias_corridos_am)
        return vna_ultimo_disponivel_ntnb * (1 + ipca_projetado_adu) ** pr1

    def ntn_b_principal(self, ytm, du, vna_ultimo_disponivel_ntnb, ipca_projetado_aa,
                        data_negociacao=DatesBR().add_working_days(DatesBR().curr_date(),
                                                                   1).strftime('%d/%m/%Y'),
                        dia_atualizacao_vna=15,
                        nper_dias_uteis_aa=252, nper_dias_corridos_am=30,
                        valor_nominal=100):
        """
        DOCSTRING: PRECIFICAÇÃO NTN-B PRINCIPAL
        INPUTS: YTM, DU, VNA ÚLTIMO DISPONÍVEL, IPCA PROJETADO AA, DATA DE NEGOCIAÇÃO (
            POR PADRÃO D+1), DIA DE ATUALIZAÇÃO DO VNA (PADRÃO 15), NÚMERO DE DIAS ÚTEIS EM UM ANO,
            NÚMERO DE DIAS CORRIDOS EM UM MÊS (PADRÃO 30) E VALOR NOMINAL DO TÍTULO (PADRÃO 100)
        OUTPUTS: FLOAT
        """
        # pr1
        pr1 = PrecificacaoTD().pr1(data_negociacao, dia_atualizacao_vna)
        # vna projetado
        vna_projetado_ntnb = PrecificacaoTD().vna_projetado_ntnb(vna_ultimo_disponivel_ntnb,
                                                                 ipca_projetado_aa, pr1,
                                                                 nper_dias_uteis_aa,
                                                                 nper_dias_corridos_am)
        # cotacao
        cotacao = valor_nominal / (1 + FinancialMath().compound_interest(ytm, nper_dias_uteis_aa,
                                                                         du)) / 100
        # pu
        return vna_projetado_ntnb * cotacao

    def ntn_b(self, ytm, lista_dus_fluxos_caixa, vna_ultimo_disponivel_ntnb, ipca_projetado_aa,
              data_negociacao=DatesBR().add_working_days(DatesBR().curr_date(),
                                                         1).strftime('%d/%m/%Y'),
              dia_atualizacao_vna=15,
              nper_dias_uteis_aa=252, nper_dias_corridos_am=30,
              valor_nominal=100, taxa_cupom_aa=0.06, periodicidade_pagmento_cupons=126):
        """
        DOCSTRING: PRECIFICAÇÃO NTN-B (COM PAGAMENTO DE CUPONS)
        INPUTS: YTM, LISTA DUS, VNA ÚLTIMO DISPONÍVEL, IPCA PROJETADO AA, DATA DE NEGOCIAÇÃO (
            POR PADRÃO D+1), DIA DE ATUALIZAÇÃO DO VNA (PADRÃO 15), NÚMERO DE DIAS ÚTEIS EM UM ANO,
            NÚMERO DE DIAS CORRIDOS EM UM MÊS (PADRÃO 30) E VALOR NOMINAL DO TÍTULO (PADRÃO 100)
        OUTPUTS: FLOAT
        """
        # pr1
        pr1 = PrecificacaoTD().pr1(data_negociacao, dia_atualizacao_vna)
        # vna projetado
        vna_projetado_ntnb = PrecificacaoTD().vna_projetado_ntnb(vna_ultimo_disponivel_ntnb,
                                                                 ipca_projetado_aa, pr1,
                                                                 nper_dias_uteis_aa,
                                                                 nper_dias_corridos_am)
        # cupom semestral
        cupom_semestral = FinancialMath().compound_interest(taxa_cupom_aa, nper_dias_uteis_aa,
                                                            periodicidade_pagmento_cupons)
        # criacao fluxo de caixa
        lista_cotacao = list()
        for du in lista_dus_fluxos_caixa:
            # cotacao
            lista_cotacao.append(cupom_semestral / (
                1 + FinancialMath().compound_interest(ytm, nper_dias_uteis_aa, du)))
        lista_cotacao.append(1 / (1 + FinancialMath().compound_interest(ytm, nper_dias_uteis_aa,
                                                                        lista_dus_fluxos_caixa[-1])))
        cotacao = sum(lista_cotacao)
        # pu
        return {
            'lista_dus_ntnb': lista_dus_fluxos_caixa,
            'lista_fluxos_caixa_ntnb': [x * vna_projetado_ntnb for x in lista_cotacao],
            'retorno_bruto_ntnb': vna_projetado_ntnb * cotacao
        }

    def vna_projetado_lft(self, vna_ultimo_disponivel_lft, taxa_selic_projetada_aa,
                          periodicidade_capitalizacao=1, nper_dias_uteis_aa=252):
        """
        DOCSTRING: VNA PROJETADO LFT
        INPUTS: VNA ÚLTIMO DISPONÍVEL (LFT), TAXA SELIC PROJETADA NOMINAL, PERIODICIDADE DE
            CAPITALIZAÇÃO (POR PADRÃO 1), DIAS ÚTEIS EM UM ANO (POR PADRÃO 252)
        OUTPUT: FLOAT
        """
        return vna_ultimo_disponivel_lft * \
            (1 + FinancialMath().compound_interest(taxa_selic_projetada_aa,
                                                   nper_dias_uteis_aa, periodicidade_capitalizacao))

    def lft(self, ytm, vna_ultimo_disponivel_lft, taxa_selic_projetada_aa,
            periodicidade_capitalizacao=1, nper_dias_uteis_aa=252, valor_nominal=1):
        """
        DOCSTRING:
        INPUTS:
        OUTPUT:
        """
        # vna atualizado
        vna_projetado_lft = PrecificacaoTD().vna_projetado_lft(vna_ultimo_disponivel_lft,
                                                               taxa_selic_projetada_aa,
                                                               periodicidade_capitalizacao,
                                                               nper_dias_uteis_aa)
        # cotacao
        cotacao = valor_nominal / (1 + FinancialMath().compound_interest(ytm, nper_dias_uteis_aa,
                                                                         periodicidade_capitalizacao))
        # pu
        return vna_projetado_lft * cotacao

    def rentabilidade_liquida(self, rentabilidade_bruta, periodo_vigencia):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        pass

    def dus_vencimento(self, data_vencimento, data_referencia=DatesBR().curr_date()):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # convertendo datas para datetime
        if type(data_vencimento) == str:
            data_vencimento = DatesBR().str_date_to_datetime(data_vencimento, 'DD/MM/AAAA')
        # lista últimos dias do ano entre o intervalo
        list_years = DatesBR().list_years_within_dates(data_referencia, data_vencimento)
        list_last_week_year_day = DatesBR().list_last_days_of_years(list_years)
        # retornando dias úteis para o vencimento
        return DatesBR().get_working_days_delta(data_referencia, data_vencimento) \
            + DatesBR().add_holidays_not_considered_anbima(data_referencia, data_vencimento,
                                                           list_last_week_year_day)

    def dus_pagamento_cupons(self, prim_dia_util, papel, vencimento):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS
        """
        # criando variáveis de passagem
        dict_cupons = dict()
        # convertendo datas para datetime
        if type(prim_dia_util) == str:
            prim_dia_util = DatesBR().str_date_to_datetime(prim_dia_util, 'DD/MM/AAAA')
        if type(vencimento) == str:
            vencimento = DatesBR().str_date_to_datetime(vencimento, 'DD/MM/AAAA')
        vencimento = DatesBR().add_working_days(
            DatesBR().sub_working_days(vencimento, 1), 1)
        # lista últimos dias do ano entre o intervalo
        list_years = DatesBR().list_years_within_dates(prim_dia_util, vencimento)
        list_last_week_year_day = DatesBR().list_last_days_of_years(list_years)
        if papel == 'ntn-f':
            # mês e dia de pagamentos de cupom
            mes_pagamento_cupom_1 = 1
            mes_pagamento_cupom_2 = 7
            dia_pagamento_cupom = 1
        # determinando a primeira data de pagamento
        if DatesBR().testing_dates(prim_dia_util, date(prim_dia_util.year,
                                                       mes_pagamento_cupom_2,
                                                       dia_pagamento_cupom)) == 'OK':
            prim_pagamento_cupom = \
                DatesBR().add_working_days(
                    DatesBR().sub_working_days(date(prim_dia_util.year,
                                                    mes_pagamento_cupom_2,
                                                    dia_pagamento_cupom), 1), 1)
        else:
            prim_pagamento_cupom = \
                DatesBR().add_working_days(
                    DatesBR().sub_working_days(date(prim_dia_util.year + 1,
                                                    mes_pagamento_cupom_1,
                                                    dia_pagamento_cupom), 1), 1)
        # * determinando o dicionário de dias de pagamentos de cupons, limitado ao vencimento
        # datas de passagem
        data_corrente = prim_dia_util
        data_prox_pagamento = prim_pagamento_cupom
        iterador = DatesBR().testing_dates(data_prox_pagamento, vencimento)
        while iterador == 'OK':
            # alterando a data corrente
            data_corrente = data_prox_pagamento
            # data próximo pagamento
            if data_prox_pagamento.month == mes_pagamento_cupom_1:
                data_prox_pagamento = \
                    DatesBR().add_working_days(DatesBR().sub_working_days(
                        date(data_prox_pagamento.year,
                             mes_pagamento_cupom_2, dia_pagamento_cupom), 1), 1)
            else:
                data_prox_pagamento = \
                    DatesBR().add_working_days(DatesBR().sub_working_days(
                        date(data_prox_pagamento.year + 1, mes_pagamento_cupom_1,
                             dia_pagamento_cupom), 1), 1)
            # preenchendo dicionário
            dict_cupons[data_corrente] = DatesBR().get_working_days_delta(
                prim_dia_util, data_corrente) \
                + DatesBR().add_holidays_not_considered_anbima(prim_dia_util, data_corrente,
                                                               list_last_week_year_day)
            # iterador
            iterador = DatesBR().testing_dates(data_prox_pagamento, vencimento)
        # resultado - dicionário de cupons
        return dict_cupons

# print(PrecificacaoTD().ltn(0.1045, 148))
# # # output
# # -886.9059241913519
# print(PrecificacaoTD().ntn_f(0.1298, [120, 248, 372, 499]))
# # output
# -953.7585159875025
# print(PrecificacaoTD().ntn_f(0.1298, [127, 251]))
# # output
# -974.6628428224586
# print(PrecificacaoTD().ntn_f(0.1298, [127, 251]) /
#       PrecificacaoTD().ntn_f(0.1298, [120, 248, 372, 499]) - 1)
# # output
# 0.021917840296620694


# ntnb-principal
# pr1
# print(PrecificacaoTD().pr1('05/01/2017'))
# # output
# 0.7096774193548387
# print(PrecificacaoTD().vna_projetado_ntnb(2494.977146,
#                                      0.0683, PrecificacaoTD().pr1('05/01/2017')))
# # output
# 2508.942630915996

# print(PrecificacaoTD().ntn_b_principal(0.0613, 1089,
#                                        PrecificacaoTD().vna_projetado_ntnb(2494.977146,
#                                                                       0.0683,
#                                                                       PrecificacaoTD().pr1(
#                                                                           '05/01/2017'))))
# # output
# 1940.139464570165


# print(PrecificacaoTD().ntn_b_principal(0.005, 1089,
#                                        PrecificacaoTD().vna_projetado_ntnb(2736.989929,
#                                                                       0.0683,
#                                                                       PrecificacaoTD().pr1(
#                                                                           '05/01/2018'))) /
#       PrecificacaoTD().ntn_b_principal(0.0613, 1089,
#                                        PrecificacaoTD().vna_projetado_ntnb(2494.977146,
#                                                                       0.0683,
#                                                                       PrecificacaoTD().pr1(
#                                                                           '05/01/2017'))) - 1)
# # output
# 0.38907055571072546

# print(PrecificacaoTD().ntn_b_principal(0.0613, 1089,
#                                        PrecificacaoTD().vna_projetado_ntnb(2494.977146,
#                                                                       0.0683,
#                                                                       PrecificacaoTD().pr1(
#                                                                           '05/01/2017'))))

# print(PrecificacaoTD().ntn_b_principal(
#     0.0613, 1089, 2494.977146, 0.0683, '05/01/2017'))
# # output
# 1940.139464570165

# print(type(DatesBR().add_working_days(DatesBR().curr_date(),
#                                       1).strftime('%d/%m/%Y')))

# print(PrecificacaoTD().ntn_b(
#     0.0610, [127, 250, 374, 500], 2508.949127, 0.0949127, '14/08/2017'))
# # # output
# # 2506.659044950116
# print(PrecificacaoTD().ntn_b(
#     0.0610, [124, 250], 2752.31, 0.0970, '14/08/2018'))
# # output
# 2751.0456728647796
# print(PrecificacaoTD().lft(0, 6543.016794, 0.1175))
# # output
# 6545.90191489939
# print(type(DatesBR().curr_date()))

# print(PrecificacaoTD().dus_pagamento_cupons(
#     DatesBR().curr_date(), 'ntn-f', '01/01/2031'))
# # output
# {datetime.date(2021, 1, 4): 3649, datetime.date(2021, 7, 1): 3471, datetime.date(2022, 1, 3): 3285, datetime.date(2022, 7, 1): 3106, datetime.date(2023, 1, 2): 2921, datetime.date(2023, 7, 3): 2739, datetime.date(2024, 1, 2): 2556, datetime.date(2024, 7, 1): 2375, datetime.date(2025, 1, 2): 2190, datetime.date(2025, 7, 1): 2010, datetime.date(2026, 1, 2): 1825, datetime.date(2026, 7, 1): 1645, datetime.date(2027, 1, 4): 1458, datetime.date(2027, 7, 1): 1280, datetime.date(2028, 1, 3): 1094, datetime.date(2028, 7, 3): 912, datetime.date(2029, 1, 2): 729, datetime.date(2029, 7, 2): 548, datetime.date(2030, 1, 2): 364, datetime.date(2030, 7, 1): 184}

# print(list(PrecificacaoTD().dus_pagamento_cupons(
#     '02/09/2020', 'ntn-f', '01/01/2031').values()))
# # output
# {datetime.date(2021, 1, 4): 3774, datetime.date(2021, 7, 1): 3774, datetime.date(2022, 1, 3): 3774, datetime.date(2022, 7, 1): 3774, datetime.date(2023, 1, 2): 3774, datetime.date(2023, 7, 3): 3774, datetime.date(2024, 1, 2): 3774, datetime.date(2024, 7, 1): 3774, datetime.date(2025, 1, 2): 3774, datetime.date(2025, 7, 1): 3774, datetime.date(
#     2026, 1, 2): 3774, datetime.date(2026, 7, 1): 3774, datetime.date(2027, 1, 4): 3774, datetime.date(2027, 7, 1): 3774, datetime.date(2028, 1, 3): 3774, datetime.date(2028, 7, 3): 3774, datetime.date(2029, 1, 2): 3774, datetime.date(2029, 7, 2): 3774, datetime.date(2030, 1, 2): 3774, datetime.date(2030, 7, 1): 3774, datetime.date(2031, 1, 2): 3774}

# int_bzd =\
#     DatesBR().get_working_days_delta(DatesBR().str_date_to_datetime('10/08/2020', 'DD/MM/AAAA'),
#                                      DatesBR().str_date_to_datetime('04/01/2021', 'DD/MM/AAAA'))
# print(int_bzd)
# print(len(DatesBR().list_rng_wds(DatesBR().str_date_to_datetime('10/08/2020', 'DD/MM/AAAA'),
#                                       DatesBR().str_date_to_datetime('04/01/2021', 'DD/MM/AAAA'))))
# print(DatesBR().list_rng_wds(DatesBR().str_date_to_datetime('10/08/2020', 'DD/MM/AAAA'),
#                                   DatesBR().str_date_to_datetime('04/01/2021', 'DD/MM/AAAA')))
# print(PrecificacaoTD().ntn_f(0.06846,
#                              list(PrecificacaoTD().dus_pagamento_cupons(
#                                  '10/08/2020',
#                                  'ntn-f', '01/01/2031').values())))
# # # output
# # -1235.9845992549463

# print(PrecificacaoTD().dus_vencimento('01/04/2021'))
# # output - 09/10/2020
# 118

# print(PrecificacaoTD().dus_vencimento('02/01/2024'))
# # output
# 808
# ! PAREI PG 62



### MULTIPROCESSING HELPER ###

import multiprocessing as mp
from typing import Callable, Optional


def mp_worker(args):
    """
    REFERENCES: https://chatgpt.com/share/6737ddcc-8564-800c-908f-9e36c311a834
    DOCSTRING: GENERALIZED WORKER FUNCTION TO CALL ANY METHOD OR FUNCTION WITH ITS ARGUMENTS
    INPUTS: ARGS (TUPLE) CONTAINING CALLABLE OBJECT, POSITIONAL ARGUMENTS AND KEYWORD ARGUMENTS:
        - CALLABLE: THE FUNCTION OR METHOD TO BE CALLED
        - POSITIONAL_ARGUMENTS: THE ARGUMENTS TO BE PASSED TO THE FUNCTION OR METHOD
        - KEYWORD_ARGUMENTS: A DICTIONARY OF KEYWORD ARGUMENTS
    OUTPUTS: THE RESULT OF THE CALLABLE
    """
    func, positional_args, keyword_args = args
    return func(*positional_args, **keyword_args)

def mp_run_parallel(
    func:Callable,
    list_task_args:Optional[list]=None,
    int_ncpus:int=mp.cpu_count() - 2 if mp.cpu_count() > 2 else 1
):
    """
    REFERENCES: https://chatgpt.com/share/6737ddcc-8564-800c-908f-9e36c311a834
    DOCSTRING: RUN WORKER PARALLIZED WITH MULTIPROCESSING, IN ORDER TO HANDLE THE PICKLING
        REQUIREMENT OF MULTIPROCESSING:
        - THE MP RELIES ON PICKLING TO SERIALIZE OBJECTS WHEN SENDING THEM TO WORKER PROCESSES
        - INSTANCE METHODS ARE NOT PICKABLE BY DEFAULT, IN ORDER TO HELP THIS A WORKER IS DEFINED
        - OBS.: ESPECIALLY ON WINDOWS, IT IS ESSENTIAL TO PROTECT THE ENTRY POINT OF THE PROGRAM
            USING IF __NAME__ == '__MAIN__' TO PREVENT RECURSIVE PROCESS SPAWNING
    INPUTS: LIST ARGS (LIST OF TUPLES, BEING THE FIRST VALUE SELF, IN CASE OF A CLASS INSTANCE),
        WORKER, INT NCPUS
    OUTPUTS: LIST
    """
    if list_task_args is None:
        list_task_args = [((), {})]
    # prepare arguments for the worker
    args_list = [(func, pos_args, kw_args) for pos_args, kw_args in list_task_args]
    # use multiprocessing Pool for parallel processing
    with mp.Pool(processes=int_ncpus) as pool:
        list_results = pool.map(mp_worker, args_list)
    # returning list of results
    return list_results




### VALIDATE BRAZILIAN DOCUMENTS NUMBERS ###

from validate_docbr import CNPJ, CPF, CNH, CNS, PIS, TituloEleitoral, RENAVAM
from requests import request
from typing import List


class DocumentsNumbersBR:

    def __init__(self, list_docs:List[str]):
        self.list_docs = list_docs

    def validate(self, doc:str = 'CNPJ') -> List[bool]:
        # determine document of interest
        if doc.upper() == 'CNPJ':
            class_ = CNPJ()
        elif doc.upper() == 'CPF':
            class_ = CPF()
        elif doc.upper() == 'CNS':
            class_ = CNS()
        elif doc.upper() == 'PIS':
            class_ = PIS()
        elif doc.upper() == 'TITULO_ELEITORAL':
            class_ = TituloEleitoral()
        elif doc.upper() == 'RENAVAM':
            class_ = RENAVAM()
        elif doc.upper() == 'CNH':
            class_ = CNH()
        else:
            raise Exception('Type of document provided not identified, please revisit '
                            + 'the variable doc. Given value: {}'.format(doc))
        # return valide/invalid reponse for each given number, provided its nature
        return class_.validate_list(self.list_docs)

    def mask_numbers(self, doc:str = 'CNPJ') -> List[str]:
        # passing variables
        list_nums_to_mask = list()
        # determine document of interest
        if doc.upper() == 'CNPJ':
            class_ = CNPJ()
        elif doc.upper() == 'CPF':
            class_ = CPF()
        elif doc.upper() == 'CNS':
            class_ = CNS()
        elif doc.upper() == 'PIS':
            class_ = PIS()
        elif doc.upper() == 'TITULO_ELEITORAL':
            class_ = TituloEleitoral()
        elif doc.upper() == 'RENAVAM':
            class_ = RENAVAM()
        elif doc.upper() == 'CNH':
            class_ = CNH()
        else:
            raise Exception('Type of document provided not identified, please revisit '
                            + 'the variable doc. Given value: {}'.format(doc))
        # iterate through each instance and mask its format, given the nature of interest
        for num in self.list_docs:
            list_nums_to_mask.append(class_.mask(num))
        # return masked value of interest
        return list_nums_to_mask

    @property
    def unmask_docs(self) -> List[str]:
        return [d.replace('.', '').replace('/', '').replace('-', '') for d in self.list_docs]

    @property
    def get_public_info_cnpj(self, url: str = 'https://receitaws.com.br/v1/cnpj/{}',
                             method: str = 'GET') -> List[dict]:
        return [request(method, url.format(cnpj_number)).json() for cnpj_number in
                self.list_docs]



### STATISTIC FUNCTIONS FOR SUPERVISED REGRESSION MACHINE LEARNING MODELS ###

import numpy as np
import statsmodels.api as sm
import statsmodels.stats.api as sms
from statsmodels.compat import lzip
from mystic.solvers import diffev2
from mystic.monitors import VerboseMonitor
from scipy.optimize import differential_evolution, curve_fit
from sklearn.linear_model import SGDRegressor, Ridge, Lasso, ElasticNet, LinearRegression, \
    LogisticRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.ensemble import RandomForestRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.metrics import confusion_matrix, classification_report, accuracy_score, \
    precision_score, recall_score, f1_score, r2_score
from sklearn.metrics import class_likelihood_ratios
from sklearn.tree import DecisionTreeRegressor
from sklearn.svm import SVR
from stpstone.transformations.cleaner.eda import ExploratoryDataAnalysis


class LinearRegressions:

    def normal_equation(self, array_x, array_y, bl_optimize=True):
        """
        REFERENCE: “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW,
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S.,
            978-1-492-03264-9.”
        DOCSTRING: NORMAL EQUATION TO FIND THE VALUE OF THETA THAT MINIMIZES THE COST FUNCTION -
            LEAST SQUARES REGRESSION
        INPUTS: ARRAY DATA AND ARRAY TARGET
        OUTPUTS: ARRAY WITH BEST THETA VECTOR, RESIDUALS, RANK
        """
        if bl_optimize == True:
            return np.linalg.lstsq(array_x, array_y, rcond=None)
        else:
            return np.linalg.inv(array_x.T.dot(array_x)).dot(array_x.T).dot(array_y)

    def batch_gradient_descent(self, array_x, array_y, max_iter=1000, eta=0.1, m=100,
                               theta=np.random.randn(2, 1)):
        """
        REFERENCE: “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW,
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S.,
            978-1-492-03264-9.”
        DOCSTRING: BATCH GRADIENT DESCENT TO FIND THE GLOBAL MINIMUM OF A LINEAR FUNCTION
        INPUTS: ARRAY DATA, ARRAY TARGET, MAX ITERATIONS, ETA (LEARNING RATE), M (ITERATIONS,
            100 AS DEFAULT), THETA (GIVEN A FIRST APROXIMATION AS DEFAULT)
        OUTPUTS: ARRAY WITH BEST THETA VECTOR
        OBSERVATIONS: 1. INCREASING THE ETA MAY IMPLY IN CONVERGING FASTER TO OPTIMAL THETA
        """
        for _ in range(max_iter):
            gradients = 2 / m * \
                array_x.T.dot(array_x.dot(theta) - array_y)
            theta = theta - eta * gradients
        return theta

    def stochastic_gradient_descent(self, array_x, array_y, method='sklearn',
                                    n_epochs=1000, t0=5, t1=50, m=100, theta=np.random.randn(2, 1),
                                    tolerance=1e-3, penalty=None, eta0=0.1):
        """
        REFERENCES: “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW,
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S.,
            978-1-492-03264-9.”
        DOCSTRING: STOCHASTIC GRADIENT DESCENT TO COVER RANDOM INSTANCES OF THE TRAINING SET AT
            EVERY STEP AND COMPUTE THE GRADIENTS BASED ONLY ON THAT SINGLE INSTANCE, AIMING TO
            FIND THE CLOSEST SOLUTION TO THE OPTIMAL THETA
        INPUTS: ARRAY DATA, ARRAY TARGET, N_EPOCHS (MAXIMUM ITERATIONS - 1000 AS DEFAULT),
            T0 AND T1 LEARNING SCHEDULE HYPERPARAMETERS, M (ITERATIONS, 100 AS DEFAULT),
            THETA (RANDOM INITIALIZATION), TOLERANCE (1E-3 AS DEFAULT), PENALTY (NONE AS DEFAULT),
            ETA0 (INITIAL LEARNING RATE, 0.1 AS DEFAULT)
        """
        if method == 'implemented':
            # defining the learning rate at each iteration, which is gradually reduced to solve the
            #   dilema of escaping from local optimal, but never settling at the minimal
            def learning_schedule(t): return t0 / (t + t1)
            # calculating the global minimun through iteration
            for epoch in range(n_epochs):
                for i in range(m):
                    random_index = np.random.randint(m)
                    xi = array_x[random_index:random_index + 1]
                    yi = array_y[random_index:random_index + 1]
                    gradients = 2 * xi.T.dot(xi.dot(theta) - yi)
                    eta = learning_schedule(epoch * m + i)
                    theta = theta - eta * gradients
            return theta
        elif method == 'sklearn':
            # defining the model
            sgd_reg = SGDRegressor(
                max_iter=n_epochs, tol=tolerance, penalty=penalty, eta0=eta0)
            # fitting the model
            sgd_reg.fit(array_x, array_y.ravel())
            # providing predictions
            array_predictions = sgd_reg.predict(array_x)
            # return desired data
            return {
                'model_fitted': sgd_reg,
                'intercept': sgd_reg.intercept_,
                'coeficients': sgd_reg.coef_,
                'predictions': array_predictions
            }

    def linear_regression(self, array_x, array_y):
        """
        DOCSTRING: BEST FITTING LINE FOR A SAMPLE OF DATA IN X ARRAY, COMAPARED TO AN Y ARRAY
        INPUTS: TWO ARRAIES '[[]]', NO NEED TO IMPORT ARRAY FUNCTION TO RECOGNISE ITS BEHAVIOUR
        OUTPUTS: DICT WIHT SCORE, COEFICIENTS, INTERCEPT, PREDICT AND THETA BEST (VECTOR WITH
            INCLINATION)
        """
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # fitting the model
        model = LinearRegression().fit(np.array(array_x),
                                      np.array(array_y))
        # retrieving the results
        if array_x.shape[1] == 1:
            return {
                'model_fitted': model,
                'score': model.score(np.array(array_x), np.array(array_y)),
                'coeficients': model.coef_,
                'intercept': model.intercept_,
                'predictions': model.predict(np.array(array_x)),
                'theta_best': np.linalg.inv(np.array(array_x).T.dot(
                    np.array(array_x))).dot(
                        np.array(array_x).T).dot(array_y)
            }
        else:
            return {
                'model_fitted': model,
                'score': model.score(np.array(array_x), np.array(array_y)),
                'coeficients': model.coef_,
                'intercept': model.intercept_,
                'predictions': model.predict(np.array(array_x))
            }

    def k_neighbors_regression(self, array_x, array_y):
        """
        DOCSTRING: BEST FITTING LINE FOR A SAMPLE OF DATA IN X ARRAY, COMAPARED TO AN Y ARRAY
        INPUTS: TWO ARRAIES '[[]]', NO NEED TO IMPORT ARRAY FUNCTION TO RECOGNISE ITS BEHAVIOUR
        OUTPUTS: DICT WIHT SCORE, COEFICIENTS, INTERCEPT, PREDICT AND THETA BEST (VECTOR WITH
            INCLINATION)
        """
        model = KNeighborsRegressor().fit(np.array(array_x),
                                      np.array(array_y))
        return {
            'model_fitted': model,
            'intercept': model.intercept_,
            'coeficients': model.coef_,
            'score': model.score(np.array(array_x), np.array(array_y)),
            'predictions': model.predict(np.array(array_x)),
            'theta_best': np.linalg.inv(np.array(array_x).T.dot(
                np.array(array_x))).dot(
                    np.array(array_x).T).dot(array_y)
        }

    def polynomial_equations(self, array_x, array_y, int_degree=2, bl_include_bias=True):
        """
        DOCSTRING: POLYNOMIAL REGRESSION TO HANDLE WITH NON-LINEAR EQUATIONS WITH A LINEAR
            APPROXIMATION
        INPUTS: ARRAY DATA, INTEGER DEGREE (2 AS DEFAULT) AND BOOLEAN TO WHETER OR NOT INCLUDE BIAS
            (FALSE AS DEFAULT)
        OUTPUTS: DICTIONARY WITH MODEL FITTED, INTERCEPT AND COEFICIENTS
        """
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # defininig polynomial model
        poly_features = PolynomialFeatures(
            degree=int_degree, include_bias=bl_include_bias)
        # transform data to polynomial purposes
        array_x_polynom = poly_features.fit_transform(array_x)
        # fit model for polynomial equation
        model = LinearRegression()
        model.fit(array_x_polynom, array_y)
        # predict based on array provided
        array_predictions = model.predict(array_x_polynom)
        # return polynomial equation
        return {
            'model_fitted': model,
            'intercept': model.intercept_,
            'coeficients': model.coef_,
            'score': model.score(np.array(array_x_polynom), np.array(array_y)),
            'predictions': array_predictions,
            'poly_features': poly_features
        }

    def ridge_regression(self, array_x, array_y, alpha=0,
                         solver_ridge_regression='cholesky'):
        """
        REFERENCES: “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW,
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S.,
            978-1-492-03264-9.”
        DOCSTRING: REGULARIZED VERSION OF LINEAR REGRESSION THAT ADDS A TERM TO THE COST
            FUNCTION, FORCING THE LEARNING ALGORITHM TO NOT ONLY FIT THE DATA, BUT ALSO KEEP
            THE MODEL WEIGHTS AS SMALL AS POSSIBLE
        INPUTS: ARRAY DATA, ARRAY TARGET, ALPHA (SENSITIVITY TO MODEL WIEGHTS - 0 AS DEFAULT),
            AND SOLVER (CHOLESKY AS DEFAULT)
        """
        # defining the model
        ridge_reg = Ridge(alpha=alpha, solver=solver_ridge_regression)
        # fit the data
        ridge_reg.fit(array_x, array_y)
        # predict based on array provided
        array_predictions = ridge_reg.predict(array_x)
        # return polynomial equation
        return {
            'model_fitted': ridge_reg,
            'intercept': ridge_reg.intercept_,
            'coeficients': ridge_reg.coef_,
            'predictions': array_predictions
        }

    def lasso_regression(self, array_x, array_y, alpha=0.1):
        """
        REFERENCES: “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW,
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S.,
            978-1-492-03264-9.”
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # declaring the model
        model = Lasso(alpha=alpha)
        # fitting curve
        model.fit(array_x, array_y)
        # making predictions
        array_predictions = model.predict(array_x)
        # return polynomial equation
        return {
            'model_fitted': model,
            'intercept': model.intercept_,
            'coeficients': model.coef_,
            'predictions': array_predictions
        }

    def elastic_net_regression(self, array_x, array_y, alpha=0.1, l1_ratio=0.5):
        """
        REFERENCES: “HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW,
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S.,
            978-1-492-03264-9.”
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # declaring the model
        model = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
        # fitting curve
        model.fit(array_x, array_y)
        # making predictions
        array_predictions = model.predict(array_x)
        # return polynomial equation
        return {
            'model_fitted': model,
            'intercept': model.intercept_,
            'coeficients': model.coef_,
            'predictions': array_predictions
        }


class NonLinearRegressions:

    def decision_tree_regression(self, array_x, array_y, seed=None):
        """
        DOCSTRING: DECISION TREE REGRESSION
        INPUTS: TWO ARRAIES '[[]]', NO NEED TO IMPORT ARRAY FUNCTION TO RECOGNISE ITS BEHAVIOUR
        OUTPUTS: DICT WIHT SCORE AND PREDICT
        """
        model = DecisionTreeRegressor(random_state=seed).fit(
            np.array(array_x),
            np.array(array_y)
        )
        return {
            'model_fitted': model,
            'score': model.score(np.array(array_x), np.array(array_y)),
            'predictions': model.predict(np.array(array_x))
        }

    def random_forest_regression(self, array_x, array_y, seed=None, n_estimators=100):
        """
        DOCSTRING: RANDOM FOREST REGRESSION, FITTING SEVERAL DECISION TREES AND AVERAGING THE
            PREDICTIONS --> BUILIDING A MODEL ONT TOP OF MANY OTHER MODELS IS CALLED ENSEMBLE
            LEARNING
        INPUTS: TWO ARRAIES '[[]]', NO NEED TO IMPORT ARRAY FUNCTION TO RECOGNISE ITS BEHAVIOUR
        OUTPUTS: DICT WIHT SCORE AND PREDICT
        """
        model = RandomForestRegressor(random_state=seed, n_estimators=n_estimators).fit(
            np.array(array_x), np.array(array_y)
        )
        return {
            'model_fitted': model,
            'score': model.score(np.array(array_x), np.array(array_y)),
            'features_importance': model.feature_importances_,
            'predictions': model.predict(np.array(array_x))
        }

    def support_vector_regression(self, array_x, array_y, kernel='poly',
                                          int_degree=2, c_positive_floating_point_number=100,
                                          epsilon=0.1):
        """
        DOCSTRING: SUPPORT VECTOR MACHINE MODEL FOR REGRESSION PURPOSES
        INPUTS: ARRAY DATA, ARRAY TARGET, KERNEL (POLY AS DEFAULT), INT DEGREE (DEGREE OF THE
            POLYNOMINAL DISTRIBUTION, BEING 2 AS DEFAULT, C (FLOATING POINT NUMBER WITH 100
            AS DEFAULT) AND EPSILON (0.1 AS DEFAULT)
        OUTPUTS: DICTIONARY WITH SCORE, PREDICT AND MODEL KEYS
        """
        model = SVR(kernel=kernel, degree=int_degree, C=c_positive_floating_point_number,
                   epsilon=epsilon)
        model = model.fit(np.array(array_x), np.array(array_y))
        return {
            'model_fitted': model,
            'score': model.score(np.array(array_x), np.array(array_y)),
            'predictions': model.predict(np.array(array_x))
        }


class LogLinearRegressions:

    def logistic_regression_logit(self, array_x, array_y,
                                  c_positive_floating_point_number=1.0, l1_ratio=None,
                                  int_max_iter=100, solver='lbfgs', penalty='l2',
                                  mult_class_classifier='auto', float_tolerance=0.0001,
                                  intercept_scaling=1, random_state=0, verbose=0,
                                  bl_fit_intercept=True, bl_warm_start=False,
                                  class_weight=None, bl_dual=False, n_jobs=None):
        """
        REFERENCE: https://realpython.com/logistic-regression-python/,
            https://www.udemy.com/course/machinelearning/
        DOCSTRING: LOGIT MODEL
        INPUTS: ARRAY DATA; ARRAY TARGET; C POSITIVE FLOATING-POINT NUMBER,
            THAT DEFINES THE RELATIVE STRENGTH OF REGULARIZATION (SMALLER VALUES INDICATE STRONGER
            REGULARIZATION, FITTING-WISE ITS POORLY FITTED, WHEREAS LARGER C MEANS WEAKER
            REGULARIZATION, THEREFORE HIGHER COEF_ AND INTERCEPT_ FOR THE MODEL); SOLVER (liblinear
            FOR LOGISTIC REGRESSION LOGIT AND lbfgs FOR SOFTMAX REGRESSION); PENALTY (l1 OR l2 -
            SCIKIT-LEARN USES L2 AS DEFAULT)
        OUTPUTS: FIT (MODEL INSTANCE), CLASSES, INTERCEPT, COEFICIENT,
            PREDICT PROBABILITY(MATRIX OF PROBABILITIES THAT THE PREDICTED OUTPUT IS EQUAL
            TO ZERO, 1-p(x), OR ONE, p(x)), SCORE (RATIO OF OBSERVATIONS CLASSIFIED CORRECTLY,
            ALSO KNOWN AS ACCURACY), CONFUSION MATRIX (PROVIDE THE ACTUAL AND PREDICTED OUTPUTS
            REGARDING TRUE NEGATIVE (C0,0), FALSE NEGATIVE (C1,0), FALSE POSITIVES (C0,1)
            AND TRUE POSITIVES (C1,1))
        """
        # checking wheter the array is unidimensional and reshaping it
        array_x = ExploratoryDataAnalysis().reshape_1d_arrays(array_x)
        # fitting the model
        model = LogisticRegression(C=c_positive_floating_point_number,
                                   class_weight=class_weight, dual=bl_dual,
                                   fit_intercept=bl_fit_intercept,
                                   intercept_scaling=intercept_scaling,
                                   l1_ratio=l1_ratio, max_iter=int_max_iter,
                                   multi_class=mult_class_classifier, n_jobs=n_jobs,
                                   penalty=penalty, random_state=random_state,
                                   solver=solver, tol=float_tolerance, verbose=verbose,
                                   warm_start=bl_warm_start).fit(array_x, array_y)
        # returning model fitted
        return {
            'model_fitted': model,
            'classes': model.classes_,
            'intercept': model.intercept_,
            'coeficient': model.coef_,
            'predict_probability': model.predict_proba(array_x),
            'predictions': model.predict(array_x),
            'score': model.score(array_x, array_y),
            'confusion_matrix': confusion_matrix(array_y,
                                                 model.predict(array_x)),
            'classification_report': classification_report(array_y,
                                                           model.predict(
                                                               array_x),
                                                           output_dict=True),
            'log_likelihood': class_likelihood_ratios(array_y, model.predict(array_x))
        }


class NonLinearEquations:

    def differential_evolution(self, cost_func, list_bounds, method='scipy', max_iter=1000,
                               max_iterations_wo_improvement=100, int_verbose_monitor=10,
                               bl_print_convergence_messages=False, bl_print_warning_messages=True,
                               bl_inter_monitor=False, int_size_trial_solution_population=40,
                               tolerance=5e-5):
        """
        REFERENCES: https://stackoverflow.com/questions/21765794/python-constrained-non-linear-optimization,
            https://mystic.readthedocs.io/en/latest/mystic.html, https://docs.scipy.org/doc/scipy/reference/optimize.html
        DOCSTRING: PRICE & STORN DIFFERENTIAL EVOLUTION SOLVER
        INPUTS:
        OUTPUTS:
        """
        if method == 'scipy':
            return differential_evolution(cost_func, list_bounds, maxiter=max_iter, tol=tolerance,
                                          disp=bl_print_convergence_messages)
        elif method == 'mystic':
            if bl_inter_monitor == True:
                mon = VerboseMonitor(int_verbose_monitor)
                return diffev2(
                    cost_func, x0=list_bounds, bounds=list_bounds,
                    npop=int_size_trial_solution_population,
                    gtol=max_iterations_wo_improvement, disp=bl_print_convergence_messages,
                    full_output=bl_print_warning_messages, itermon=mon, maxiter=max_iter,
                    ftol=tolerance)
            else:
                return diffev2(
                    cost_func, x0=list_bounds, bounds=list_bounds,
                    npop=int_size_trial_solution_population,
                    gtol=max_iterations_wo_improvement, disp=bl_print_convergence_messages,
                    full_output=bl_print_warning_messages, maxiter=max_iter, ftol=tolerance)
        else:
            raise Exception(
                'Method not recognized, please revisit the parameter')

    def optimize_curve_fit(self, func, array_x, array_y):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return curve_fit(func, xdata=array_x, y_data=array_y)

    def polynomial_fit(self, array_x, array_y):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS: ARRAY ESTIMATIVES AND ARRAY POLYNOMIAL INTERPOLATED VALUES
        """
        # estimatives
        y_est = np.polyfit(array_x, array_y)
        # return estimatives and values interpolated
        return {
            'coefficients': y_est,
            'values_interpolated': np.polyval(y_est, array_x)
        }



from typing import Optional, Union, Literal


class LiquidityRatios:
    """
    Metadata: https://www.investopedia.com/terms/l/liquidityratios.asp,
    https://www.toppr.com/guides/principles-and-practice-of-accounting/accounting-ratios/liquidity-ratios/
    """

    def current_ratio(self, float_curr_assets: float, float_curr_liabilities: float):
        """
        Calculates the current ratio, which is a liquidity ratio that measures a company's ability
        to pay short-term obligations or those due within one year. The ratio indicates the extent
        to which current assets cover current liabilities.

        Args:
            float_curr_assets (float): Total current assets.
            float_curr_liabilities (float): Total current liabilities.

        Returns:
            float: The current ratio, calculated as current assets divided by current liabilities.
        """
        return float_curr_assets / float_curr_liabilities

    def quick_ratio(self, float_curr_assets: float, float_inventories: float,
                    float_curr_liabilities: float):
        """
        Calculates the quick ratio, which is a liquidity ratio that measures a company's ability
        to pay short-term obligations or those due within one year. The ratio indicates the extent
        to which current assets, excluding inventory, cover current liabilities.

        Args:
            float_curr_assets (float): Total current assets.
            float_inventories (float): Inventories.
            float_curr_liabilities (float): Total current liabilities.

        Returns:
            float: The quick ratio, calculated as current assets minus inventories divided by
            current liabilities.
        """
        return (float_curr_assets - float_inventories) / float_curr_liabilities

    def dso(self, float_avg_accounts_rec:float, float_rev:float,
            int_wdy: Optional[Union[int, Literal[252, 365, 360]]]=365):
        """
        Calculates the days sales outstanding (DSO), which is the average number of days that
        a company takes to collect its accounts receivable. The DSO is used to measure the
        efficiency of a company's accounts receivable management, and it is often used to
        evaluate the creditworthiness of a company.

        Args:
            float_avg_accounts_rec (float): Average accounts receivable.
            float_rev (float): Total revenue.
            int_wdy (Optional[Union[int, Literal[252, 365, 360]]], optional): Number of days in
                the period. Defaults to 365.

        Returns:
            float: The DSO, calculated as the number of days divided by the revenue minus the
            average accounts receivable.
        """
        return float(int_wdy) / (float_rev / float_avg_accounts_rec)

    def cash_ratio(self, float_cash_eq: float, float_curr_liabilities: float):
        """
        Calculates the cash ratio, which is a liquidity ratio that measures a company's ability
        to pay off its short-term liabilities with its most liquid assets, cash and cash equivalents.

        Args:
            float_cash_eq (float): Total cash and cash equivalents.
            float_curr_liabilities (float): Total current liabilities.

        Returns:
            float: The cash ratio, calculated as cash and cash equivalents divided by current liabilities.
        """
        return float_cash_eq / float_curr_liabilities


class SolvencyRatios:
    """
    Metadata: https://www.investopedia.com/terms/s/solvencyratio.asp,
        https://medium.com/quant-factory/calculating-altman-z-score-with-python-3c6697ee7aee
    """

    def interest_coverage_ratio(self, float_ebit: float, float_int_exp: float):
        """
        Calculates the interest coverage ratio, which is a solvency ratio that measures a company's
        ability to pay interest on its debt. The interest coverage ratio is calculated by dividing
        earnings before interest and taxes (EBIT) by interest expenses.

        Args:
            float_ebit (float): Earnings before interest and taxes.
            float_int_exp (float): Interest expenses.

        Returns:
            float: The interest coverage ratio, calculated as EBIT divided by interest expenses.
        """
        return float_ebit / float_int_exp

    def debt_to_assets_ratio(self, float_debt: float, float_assets: float):
        return float_debt / float_assets

    def equiy_ratio(self, float_tse: float, float_assets: float):
        """
        Calculates the equity ratio, which is a solvency ratio that measures the proportion of a
        company's total assets financed by shareholders' equity. The equity ratio is an indicator of
        financial stability, showing the percentage of assets owned by shareholders.

        Args:
            float_tse (float): Total shareholders' equity.
            float_assets (float): Total assets.

        Returns:
            float: The equity ratio, calculated as total shareholders' equity divided by total assets.
        """
        return float_tse / float_assets

    def debt_to_equity_ratio(self, float_debt: float, float_equity: float):
        return float_debt / float_equity

    def altmans_z_score(self, float_nwk_cap: float, float_total_assets: float,
                        float_ret_earnings: float, float_ebit: float,
                        float_mkt_cap: float, float_total_liabilities: float, float_sales: float):
        """
        Calculates the Altman Z-score, which is a credit-strength test that assesses a company's
        likelihood of bankruptcy. The Z-score is calculated from five financial ratios: working
        capital to total assets, retained earnings to total assets, earnings before interest and
        taxes (EBIT) to total assets, market capitalization to total liabilities, and sales to total
        assets.

        Args:
            float_nwk_cap (float): Working capital.
            float_total_assets (float): Total assets.
            float_ret_earnings (float): Retained earnings.
            float_ebit (float): Earnings before interest and taxes.
            float_mkt_cap (float): Market capitalization.
            float_total_liabilities (float): Total liabilities.
            float_sales (float): Sales.

        Returns:
            float: The Altman Z-score.
        """
        return 1.2 * (float_nwk_cap / float_total_assets) \
            + 1.4 * (float_ret_earnings / float_total_assets) \
            + 3.3 * (float_ebit / float_total_assets) \
            + 0.6 * (float_mkt_cap / float_total_liabilities) \
            + 1.0 * (float_sales / float_total_assets)



import re
from typing import Union, Dict, List, Optional
from logging import Logger
from requests import request
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD
from stpstone.utils.connections.netops.proxies.abc import ABCSession
from stpstone.utils.geography.ww import WWTimezones, WWGeography
from stpstone.utils.parsers.numbers import NumHandler
from stpstone.utils.parsers.str import StrHandler


class SpysOneCountry(ABCSession):

    def __init__(
        self,
        bl_new_proxy: bool = True,
        dict_proxies: Union[Dict[str, str], None] = None,
        int_retries: int = 10,
        int_backoff_factor: int = 1,
        bl_alive: bool = True,
        list_anonymity_value: List[str] = ["anonymous", "elite"],
        list_protocol: str = 'http',
        str_continent_code: Union[str, None] = None,
        str_country_code: Union[str, None] = None,
        bl_ssl: Union[bool, None] = None,
        float_min_ratio_times_alive_dead: Optional[float] = 0.02,
        float_max_timeout: Optional[float] = 600,
        bl_use_timer: bool = False,
        list_status_forcelist: List[int] = [429, 500, 502, 503, 504],
        logger: Optional[Logger] = None,
        int_wait_load_seconds: int = 10,
    ) -> None:
        super().__init__(
            bl_new_proxy=bl_new_proxy,
            dict_proxies=dict_proxies,
            int_retries=int_retries,
            int_backoff_factor=int_backoff_factor,
            bl_alive=bl_alive,
            list_anonymity_value=list_anonymity_value,
            list_protocol=list_protocol,
            str_continent_code=str_continent_code,
            str_country_code=str_country_code,
            bl_ssl=bl_ssl,
            float_min_ratio_times_alive_dead=float_min_ratio_times_alive_dead,
            float_max_timeout=float_max_timeout,
            bl_use_timer=bl_use_timer,
            list_status_forcelist=list_status_forcelist,
            logger=logger
        )
        self.int_wait_load_seconds = int_wait_load_seconds
        self.fstr_url = "https://spys.one/free-proxy-list/{}/"
        self.xpath_tr = '//tr[contains(@class, "spy1x")]'
        self.xpath_dd_anonimity = '//select[@name="xf1"]'
        self.xpath_dd_show = '//select[@name="xpp"]'
        self.xpath_ssl = '//select[@name="xf2"]'
        self.xpath_type = '//select[@name="xf5"]'
        self.ww_timezones = WWTimezones()
        self.ww_geography = WWGeography()
        self.num_handler = NumHandler()
        self.str_handler = StrHandler()

    def parse_location(self, location_str: str) -> Optional[Dict[str, str]]:
        """
        Parses location strings in the format:
        "BR São João da Ponte (Minas Gerais)" or "BR Birigui (São Paulo)"

        Args:
            location_str (str): The location string to parse.

        Returns:
            Optional[Dict[str, str]]: A dictionary containing country, city, and state.
        """
        if len(location_str) == 2: return location_str, "N/A", "N/A"
        elif self.str_handler.match_string_like(location_str, "*(*") == False:
            re_pattern = r'(?i)^([a-z]{2})\s*([^*]+?)$'
            re_match = re.match(re_pattern, location_str.strip())
            if re_match is not None:
                return re_match.group(1), re_match.group(2), "N/A"
            return None
        re_pattern = r'(?i)^([a-z]{2})\s*([^*]+?)\(([^*]+?)\)'
        re_match = re.match(re_pattern, location_str.strip())
        if re_match is not None:
            return re_match.group(1), re_match.group(2), re_match.group(3)
        return None

    @property
    def _available_proxies(self) -> List[Dict[str, Union[str, float]]]:
        list_ser = list()
        cls_selenium_wd = SeleniumWD(
            url=self.fstr_url.format(self.str_country_code.upper()),
            bl_headless=True,
            bl_incognito=True,
            int_wait_load_seconds=self.int_wait_load_seconds
        )
        try:
            web_driver = cls_selenium_wd.get_web_driver
            cls_selenium_wd.wait_until_el_loaded(self.xpath_tr)
            for el_, el_opt in [
                (self.xpath_dd_anonimity, './option[@value="1"]'),
                (self.xpath_dd_show, './/option[last()]'),
                (self.xpath_ssl, './/option[@value="1"]'),
                (self.xpath_type, './/option[@value="1"]'),
            ]:
                el_ = cls_selenium_wd.find_element(web_driver, self.xpath_dd_anonimity)
                el_opt = cls_selenium_wd.find_element(el_, './option[@value="1"]')
                el_opt.click()
                cls_selenium_wd.wait(10)
            el_trs = cls_selenium_wd.find_elements(web_driver, self.xpath_tr)
            for el_tr in el_trs:
                try:
                    tup_ip_port = cls_selenium_wd.find_element(el_tr, './td[1]').text.split(":")
                    str_ip = tup_ip_port[0]
                    str_port = tup_ip_port[1]
                except IndexError:
                    continue
                if self.num_handler.is_numeric(str_port) == False: continue
                str_protocol = cls_selenium_wd.find_element(el_tr, './td[2]').text.lower()
                str_protocol = str_protocol.split(" ")[0]
                str_anonimity = cls_selenium_wd.find_element(el_tr, './td[3]').text.lower()
                str_anonimity = "elite" if str_anonimity == "hia" else (
                    "anonymous" if str_anonimity == "anm" else "transparent"
                )
                str_country_city_region = cls_selenium_wd.find_element(el_tr, './td[4]').text
                str_country_code, str_city, str_region = self.parse_location(str_country_city_region)
                str_org_name = cls_selenium_wd.find_element(el_tr, './td[5]').text
                str_latency = cls_selenium_wd.find_element(el_tr, './td[6]').text
                float_latency = float(str_latency.replace(".", ""))
                str_uptime = cls_selenium_wd.find_element(el_tr, './td[8]').text
                str_check_date = cls_selenium_wd.find_element(el_tr, './td[9]').text
                ts_check_date = self.composed_time_ago_to_ts_unix(str_check_date)
                list_ser.append({
                    "protocol": str_protocol,
                    "bl_alive": True,
                    "status": "success",
                    "alive_since": ts_check_date,
                    "anonymity": str_anonimity,
                    "average_timeout": 1.0 / float_latency,
                    "first_seen": ts_check_date,
                    "ip_data": "",
                    "ip_name": "",
                    "timezone": ", ".join(self.ww_timezones.get_timezones_by_country_code(
                        self.str_country_code)),
                    "continent": self.ww_geography.get_continent_by_country_code(
                        self.str_country_code),
                    "continent_code": self.ww_geography.get_continent_code_by_country_code(
                        self.str_country_code),
                    "country": self.ww_geography.get_country_details(self.str_country_code)["name"],
                    "country_code": self.str_country_code,
                    "city": str_city,
                    "district": str_region,
                    "region_name": str_region,
                    "zip": "",
                    "bl_hosting": False,
                    "isp": "",
                    "latitude": 0.0,
                    "longitude": 0.0,
                    "organization": str_org_name,
                    "proxy": f"{str_ip}:{str_port}",
                    "ip": str_ip,
                    "port": str_port,
                    "bl_ssl": True,
                    "timeout": 1.0 / float_latency,
                    "times_alive": 0,
                    "times_dead": 0,
                    "ratio_times_alive_dead": 1.0,
                    "uptime": str_uptime
                })
        finally:
            web_driver.quit()
        return list_ser



import os
from pathlib import Path
from datetime import datetime
from playwright.sync_api import sync_playwright
from typing import Optional, List, Dict, Union
from contextlib import contextmanager
from logging import Logger
from stpstone.utils.loggs.create_logs import CreateLog


class PlaywrightScraper:
    def __init__(
        self,
        bl_headless: bool = True,
        user_agent: Optional[str] = None,
        proxy: Optional[str] = None,
        viewport: Dict[str, int] = None,
        int_default_timeout: int = 30000,
        bl_accept_cookies: bool = True, 
        bl_incognito: bool = False,
        logger: Optional[Logger] = None
    ):
        """
        Initialize a generic Playwright scraper
        
        Args:
            bl_headless (bool): Run in bl_headless mode
            user_agent (str): Custom user agent string
            proxy (str): Proxy server address
            viewport (dict): Browser viewport settings {"width": 1920, "height": 1080}
            int_default_timeout (int): Default timeout in milliseconds
            bl_accept_cookies (bool): Attempt to accept cookies if popup appears
            bl_incognito (bool): Run in incognito mode
        """
        self.bl_headless = bl_headless
        self.user_agent = user_agent or "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
        self.proxy = proxy
        self.viewport = viewport or {"width": 1920, "height": 1080}
        self.int_default_timeout = int_default_timeout
        self.bl_accept_cookies = bl_accept_cookies
        self.bl_incognito = bl_incognito
        self.logger = logger
        self.playwright = None
        self.browser = None
        self.context = None
        self.page = None

    @contextmanager
    def launch(self):
        """Context manager for browser session"""
        try:
            self.playwright = sync_playwright().start()
            browser_args = {
                "headless": self.bl_headless,
                "proxy": {"server": self.proxy} if self.proxy else None
            }
            if self.bl_incognito:
                self.context = self.playwright.chromium.launch_persistent_context(
                    user_data_dir=None,
                    **browser_args,
                    viewport=self.viewport,
                    user_agent=self.user_agent
                )
                self.page = self.context.pages[0]
            else:
                self.browser = self.playwright.chromium.launch(**browser_args)
                self.context = self.browser.new_context(
                    viewport=self.viewport,
                    user_agent=self.user_agent
                )
                self.page = self.context.new_page()
            self.page.set_default_timeout(self.int_default_timeout)
            yield self
        except Exception as e:
            CreateLog().log_message(self.logger, f"Error launching browser: {e}", "error")
            raise
        finally:
            self.close()

    def close(self):
        """Clean up resources"""
        if hasattr(self, "context") and self.context:
            self.context.close()
        if hasattr(self, "browser") and self.browser:
            self.browser.close()
        if hasattr(self, "playwright") and self.playwright:
            self.playwright.stop()

    def navigate(self, url: str, timeout: Optional[int] = None):
        """Navigate to a URL"""
        try:
            self.page.goto(url, timeout=timeout or self.int_default_timeout)
            if self.bl_accept_cookies:
                self._handle_cookie_popup()
            return True
        except Exception as e:
            CreateLog().log_message(self.logger, f"Error navigating to {url}: {e}", "error")
            return False

    def get_current_url(self) -> Optional[str]:
        """
        Get the current page URL
        
        Returns:
            str: Current URL if page exists, None otherwise
        """
        if hasattr(self, 'page') and self.page:
            return self.page.url
        return None

    def _handle_cookie_popup(self, timeout: int = 3000):
        """Attempt to accept cookies if popup appears"""
        try:
            self.page.click("text=Accept All", timeout=timeout)
            CreateLog().log_message(self.logger, "Accepted cookies", "info")
        except:
            pass

    def selector_exists(
        self,
        selector: str,
        selector_type: str = "xpath",
        timeout: Optional[int] = None,
        visible: Optional[bool] = None
    ) -> bool:
        """
        Check if a selector exists on the page
        
        Args:
            selector (str): The selector to check
            selector_type (str): "xpath" or "css"
            timeout (int): Maximum time to wait in milliseconds (None for no wait)
            visible (bool): Check for visibility (True=visible, False=hidden, None=either)
            
        Returns:
            bool: True if selector exists (with given visibility), False otherwise
        """
        try:
            if selector_type.lower() not in ("xpath", "css"):
                raise ValueError(f"Unsupported selector type: {selector_type}")
            if timeout is not None:
                # wait for selector with specified visibility
                state = "visible" if visible is True else (
                        "hidden" if visible is False else "attached")
                self.page.wait_for_selector(
                    selector,
                    state=state,
                    timeout=timeout
                )
                return True
            else:
                # immediate check without waiting
                if visible is True:
                    return self.page.locator(selector).first.is_visible()
                elif visible is False:
                    return self.page.locator(selector).first.is_hidden()
                else:
                    return self.page.locator(selector).count() > 0
                    
        except Exception as e:
            CreateLog().log_message(
                self.logger, 
                f"Selector check failed for {selector}: {str(e)}", 
                "warning"
            )
            return False

    def get_element(
        self,
        selector: str,
        selector_type: str = "xpath",
        timeout: Optional[int] = None,
        visible: bool = True
    ) -> Optional[Dict[str, Union[str, Dict]]]:
        """
        Get a single element matching selector
        
        Args:
            selector (str): The selector to find the element
            selector_type (str): "xpath" or "css"
            timeout (int): Maximum time to wait in milliseconds
            visible (bool): Wait for element to be visible
            
        Returns:
            Dictionary with element"s text, html and bounding box info, or None if not found
        """
        try:
            if selector_type.lower() == "xpath":
                if timeout is not None:
                    self.page.wait_for_selector(
                        selector,
                        state="visible" if visible else "attached",
                        timeout=timeout or self.int_default_timeout
                    )
                element = self.page.locator(selector).first
            elif selector_type.lower() == "css":
                if timeout is not None:
                    self.page.wait_for_selector(
                        selector,
                        state="visible" if visible else "attached",
                        timeout=timeout or self.int_default_timeout
                    )
                element = self.page.locator(selector).first
            else:
                raise ValueError(f"Unsupported selector type: {selector_type}")
            if not element.text_content().strip():
                return None
            return {
                "text": element.text_content().strip(),
                "html": element.inner_html(),
                "bounding_box": element.bounding_box()
            }
        except Exception as e:
            CreateLog().log_message(self.logger, f"Error getting element: {e}", "error")
            return None
    
    def get_element_attrb(
        self,
        selector: str,
        str_attribute: str = "href",
        selector_type: str = "xpath"
    ) -> Optional[str]:
        try:
            if selector_type == "xpath":
                element = self.page.locator(selector).first
            elif selector_type == "css":
                element = self.page.locator(selector).first
            else:
                raise ValueError(f"Unsupported selector type: {selector_type}")
            return element.get_attribute(str_attribute)
        except Exception as e:
            CreateLog().log_message(self.logger, f"Error getting href: {e}", "error")
            return None

    def get_elements(
        self,
        selector: str,
        selector_type: str = "xpath",
        timeout: Optional[int] = None,
        visible: bool = True
    ) -> List[Dict[str, Union[str, Dict]]]:
        """
        Get elements matching selector
        
        Args:
            selector (str): The selector to find elements
            selector_type (str): "xpath" or "css"
            timeout (int): Maximum time to wait in milliseconds
            visible (bool): Wait for elements to be visible
            
        Returns:
            List of elements with text, html and bounding box info
        """
        try:
            if selector_type.lower() == "xpath":
                self.page.wait_for_selector(
                    selector,
                    state="visible" if visible else "attached",
                    timeout=timeout or self.int_default_timeout
                )
                elements = self.page.locator(selector).all()
            elif selector_type.lower() == "css":
                self.page.wait_for_selector(
                    selector,
                    state="visible" if visible else "attached",
                    timeout=timeout or self.int_default_timeout
                )
                elements = self.page.locator(selector).all()
            else:
                raise ValueError(f"Unsupported selector type: {selector_type}")
            return [{
                "text": el.text_content().strip(),
                "html": el.inner_html(),
                "bounding_box": el.bounding_box()
            } for el in elements if el.text_content().strip()]
        except Exception as e:
            CreateLog().log_message(self.logger, f"Error getting elements: {e}", "error")
            return []

    def get_list_data(
        self,
        table_selector: str,
        selector_type: str = "xpath",
        timeout: Optional[int] = None
    ) -> List[str]:
        """
        Get text content from table cells
        
        Args:
            table_selector (str): Selector for table or table cells
            selector_type (str): "xpath" or "css"
            timeout (int): Maximum time to wait in milliseconds
            
        Returns:
            List of text content from table cells
        """
        elements = self.get_elements(table_selector, selector_type, timeout)
        return [el["text"] for el in elements]
    
    def export_html(
        self,
        content: str,
        folder_path: str = "scraped_data",
        filename: Optional[str] = None,
        bl_include_timestamp: bool = True
    ) -> str:
        """
        Export HTML content to a file in the specified folder.
        
        Args:
            content (str): HTML content to save
            folder_path (str): Path to the output folder (default: "scraped_data")
            filename (str): Optional custom filename (without extension)
            bl_include_timestamp (bool): Whether to append bl_include_timestamp to filename
            
        Returns:
            str: Path to the saved file
        """
        try:
            Path(folder_path).mkdir(parents=True, exist_ok=True)
            if not filename:
                url = self.page.url if hasattr(self, "page") and self.page else "scraped"
                filename = url.split("//")[-1].replace("/", "_").replace("?", "_").replace("=", "_")
            if bl_include_timestamp:
                timestamp_str = datetime.now().strftime("%Y%m%d_%H%M%S")
                filename = f"{filename}_{timestamp_str}"
            if not filename.endswith(".html"):
                filename += ".html"
            file_path = os.path.join(folder_path, filename)
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(content)
            CreateLog().log_message(self.logger, f"HTML content saved to {file_path}", "info")
            return file_path
        except Exception as e:
            CreateLog().log_message(self.logger, f"Error saving HTML file: {e}", "error")
            raise Exception(f"Error saving HTML file: {e}")


import pandas as pd
from datetime import datetime
from typing import Optional, List, Any, Dict, Tuple
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from bs4 import BeautifulSoup
from math import nan
from stpstone._config.global_slots import YAML_ANBIMA_FORECASTS
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.parsers.html import HtmlHandler
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.loggs.create_logs import CreateLog


class AnbimaForecasts(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None
    ) -> None:
        super().__init__(
            dict_metadata=YAML_ANBIMA_FORECASTS,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs

    def td_parser(self, root: BeautifulSoup, source: str) -> Dict[str, List[Any]]:
        list_data = list()
        for i_tb, bs_table in enumerate(root.find_all('table')):
            if source.lower() == 'mp1':
                if (i_tb + 1) % 3 != 1: continue
            elif source.lower() == 'mp2':
                if (i_tb + 1)  % 3 != 2: continue
            elif source.lower() == 'ltm':
                if (i_tb + 1)  % 3 != 0: continue
            else:
                raise ValueError("Invalid source")
            if i_tb // 3 == 0:
                str_pmi = 'IGPM'
            elif i_tb // 3 == 1:
                str_pmi = 'IPCA'
            else:
                raise ValueError("Invalid index")
            for i_tr, bs_tr in enumerate(bs_table.find_all('tr')):
                if i_tr >= 2:
                    list_ = [
                        x.get_text().replace("-", "9999.99") if len(x.get_text()) == 1
                        else x.get_text().replace(",", ".")
                        for x in bs_tr.find_all('td')
                    ]
                    if (source.lower() == "ltm") and (len(list_) == 3):
                        list_.insert(0, nan)
                        list_.insert(len(list_), nan)
                    list_.extend([str_pmi, source.upper()])
                    list_data.extend(list_)
        return list_data

    def table_(self, list_data: List[Any], source: str) -> pd.DataFrame:
        list_cols = list(YAML_ANBIMA_FORECASTS[source]["dtypes"].keys())
        list_ser = HandlingDicts().pair_headers_with_data(list_cols, list_data)
        df_ = pd.DataFrame(list_ser)
        df_["PROJECAO"] = df_["PROJECAO"].astype(float) / 100.0
        if "VALOR_EFETIVO" in list(df_.columns):
            df_["VALOR_EFETIVO"] = df_["VALOR_EFETIVO"].astype(float) / 100.0
        return df_

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        root = HtmlHandler().bs_parser(resp_req)
        source = self.get_query_params(resp_req.url, "source")
        list_data = self.td_parser(root, source)
        return self.table_(list_data, source)




### METHODS TO HANDLE STRINGS ###

import json
import re
import uuid
import ftfy
from base64 import b64encode
from fnmatch import fnmatch
from string import ascii_lowercase, ascii_uppercase, digits
from typing import Any, Dict, Union, List
from unicodedata import combining, normalize
from urllib.parse import urlparse, parse_qs
from basicauth import encode
from bs4 import BeautifulSoup
from unidecode import unidecode


class StrHandler:

    @property
    def multi_map_reference(self):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: TRANSFORM SOME WESTERN TYPOGRAPHICAL SYMBOLS INTO ASCII, BUILDING MAPPING TABLE
            FOR CHAR-TO-CHAR REPLACEMENT
        INPUTS: -
        OUTPUT: DICT
        """
        single_map = str.maketrans("""‚ƒ„†ˆ‹‘’“”•–—˜›""', '""'f"*^<''""---~>""")
        multi_map = str.maketrans(
            {
                "€": "<euro>",
                "…": "...",
                "OE": "OE",
                "™": "(TM)",
                "oe": "oe",
                "‰": "<per mille>",
                "‡": "**",
            }
        )
        return multi_map.update(single_map)

    def get_between(self, s, first, last):
        """
        DOCSTRING: FIND STRINGS BETWEEN TWO SUBSTRINGS
        INPUTS: ORIGINAL STRING, INITAL AND FINAL DELIMITERS
        OUTPUTS: MID STRING
        """
        try:
            start = s.index(first) + len(first)
            end = s.index(last, start)
            return s[start:end]
        except ValueError:
            return ""
    
    def get_after(self, s, first):
        """
        DOCSTRING: FIND STRINGS BETWEEN TWO SUBSTRINGS
        INPUTS: ORIGINAL STRING, INITAL AND FINAL DELIMITERS
        OUTPUTS: MID STRING
        """
        try:
            start = s.index(first) + len(first)
            return s[start:]
        except ValueError:
            return ""

    def find_substr_str(self, str_, substr_):
        """
        DOCSTRING: FINDING A SUBSTRING IN A STRING
        INPUTS: STRING AND SUBSTRING
        OUTPUTS: BOOLEAN
        """
        return substr_ in str_

    def match_string_like(self, str_, str_like):
        """
        DOCSTRING: MATCHING STRING WITH RESPECTIVELY STRING LIKE
        INPUTS: STRING AND STRING LIKE
        OUTPUTS: BOOLEAN
        """
        return fnmatch(str_, str_like)

    def latin_characters(self, str_):
        """
        DOCSTRING: CORRECTING SPECIAL CHARACTERS
        INPUTS: STRING
        OUTPUTS: CORRECTED STRING
        """
        return str_.encode("latin1").decode("utf-8")

    def decode_special_characters_ftfy(self, str_):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return ftfy.fix_str_(str_)

    def removing_accents(self, str_):
        """
        DOCSTRING: REMOVE ACCENTS FROM LATIN ALPHABET
        INPUTS: STRING
        OUTPUTS: STRING
        """
        return unidecode(str_)

    def byte_to_latin_characters(self, str_):
        """
        DOCSTRING: CORRECTING SPECIAL CHARACTERS
        INPUTS: STRING
        OUTPUTS: CORRECTED STRING
        """
        return str_.encode("latin1").decode("ISO-8859-1")

    def remove_diacritics(self, str_):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015
        LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSITRNGS: REMOVE ALL DIACRITICS FROM A STRING, SUCH AS ACCENTS, CEDILLAS, ETC, FROM LATIN
            AND NON-LATIN ALPHABET, LIKE GREEK.
        INPUTS: STRING
        OUTPUTS: STRING
        """
        norm_txt = normalize("NFD", str_)
        shaved = "".join(c for c in norm_txt if not combining(c))
        return normalize("NFC", shaved)

    def remove_diacritics_nfkd(self, str_: str, bl_lower_case: bool = True) -> str:
        if bl_lower_case == True:
            str_ = str_.lower()
        str_ = str_.replace("\n", "")
        return "".join(c for c in normalize("NFKD", str_) if not combining(c))

    def normalize_text(self, str_):
        return normalize("NFKD", str_).encode("ascii", "ignore").decode("utf-8")

    def remove_sup_period_marks(self, corpus, patterns=r"[!.?+]"):
        """
        DOCSTRING: REMOVE END PERIOD MARKS
        INPUTS: CORPUS AND PATTERNS (DEFAULT)
        OUTPUTS: STRING
        """
        return re.sub(patterns, "", corpus)

    def remove_only_latin_diacritics(self, str_, latin_base=False):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015
            LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: REMOVE ALL DISCRITIC MARKS FROM LATIN BASE CHARACTERS
        INPUTS: STRING, LATIN BASE (FALSE AS DEFAULT)
        OUTPUTS: STRING
        """
        norm_txt = normalize("NFD", str_)
        keepers = []
        for c in norm_txt:
            if combining(c) and latin_base:
                continue  # ignore diacritic on Latin base char
            keepers.append(c)
            # if it isn't combining char, it's a new base char
            if not combining(c):
                latin_base = c in str_.ascii_letters
        shaved = "".join(keepers)
        return normalize("NFC", shaved)

    def dewinize(self, str_):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: REPLACE WIN1252 SYMBOLS WITH ASCII CHARS OR SEQUENCES
        INPUTS: STRING
        OUTPUTS: STRING
        """
        return str_.translate(self.multi_map_reference)

    def asciize(self, str_):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: APPLY NFKC NORMALIZATION TO COMPOSE CHARACTERS WITH THEIR COMPATIBILITY CODE
            POINTS IN ASCII SYSTEM
        INPUTS: STRING
        OUTPUTS: STRING
        """
        no_marks = self.remove_only_latin_diacritics(self.dewinize(str_))
        no_marks = no_marks.replace("ß", "ss")
        return normalize("NFKC", no_marks)

    def remove_substr(self, str_, substr_):
        """
        DOCSTRING: REMOVE A SUBSTRING FROM A GIVEN STRING
        INPUTS: STRING AND SUBSTRING
        OUTPUTS: STRING WITHOUT SUBSTRING
        """
        return str_.replace(substr_, "")

    def get_string_until_substr(self, str_, substring):
        """
        DOCSTRING: RETURN A STRING UNTIL FIND ITS SUBSTRING
        INPUTS: STRING, SUBSTRING
        OUTPUTS: STRING
        """
        return str_.split(substring)[0]

    def get_string_after_substr(self, str_, substring):
        """
        DOCSTRING: RETURN A STRING AFTER FIND ITS SUBSTRING
        INPUTS: STRING, SUBSTRING
        OUTPUTS: STRING
        """
        return str_.split(substring)[1]

    def base64_encode(self, userid, password):
        """
        DOCSTRING: ENCODING IN BASE 64 AN USER AND PASSWORD COMBINATION
        INPUTS: STRING TO ENCODE
        OUTPUTS: STRING ENCODED IN BASE64
        """
        return encode(userid, password)

    def base64_str_encode(self, str_, code_method="ascii"):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # encode to bytes
        message_bytes = str_.encode(code_method)
        # encode bytes to base64
        base64_bytes = b64encode(message_bytes)
        base64_message = base64_bytes.decode(code_method)
        # return message
        return base64_message

    @property
    def universally_unique_identifier(self):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # make a new uuid
        uuid_identifier = uuid.uuid4()
        # return uudi
        return {
            "uuid": uuid_identifier,
            "uuid_hex_digits_str": str(uuid_identifier),
            "uuid_32_character_hexadecimal_str": uuid_identifier.hex,
        }

    def letters_to_numbers(
        self,
        letters_in_alphabet=21,
        first_letter_alphabet="f",
        list_not_in_range=["i", "l", "o", "p", "r", "s", "t", "w", "y"],
    ):
        """
        DOCSTRING: JSON CORRELATING LETTERS AND NUMBERS
        INPUTS: LETTERS IN ALPHABET FROM THE FIRST ONE (21 AS DEFAULT),
            FIRST LETTER IN ALPHABET (F AS DEFAULT), LIST NOT IN RANGE (I, L, O, P, R, S, T, W, Y
            AS DEFAULT)
        OUTPUTS: JSON WITH LETTERS IN LOWER CASE AS KEYS
        """
        # auxiliary variables
        dict_message = dict()
        i_aux = 0

        # dictionary correlating letters and numbers
        for i in range(
            ord(first_letter_alphabet), ord(first_letter_alphabet) + letters_in_alphabet
        ):
            if chr(i) not in list_not_in_range:
                dict_message[chr(i)] = i - 101 - i_aux
            else:
                i_aux += 1

        # json to export
        return json.loads(json.dumps(dict_message))

    def alphabetic_range(self, case="upper"):
        """
        DOCSTRING: ALPHABETIC RANGE IN UPPER OR LOWER CASE
        INPUTS: CASE
        OUTPUTS: LIST
        """
        if case == "upper":
            return list(ascii_uppercase)
        elif case == "lower":
            return list(ascii_lowercase)
        else:
            raise Exception(
                "Case ought be upper or lower, although {} was given, ".format(case)
                + "please revisit the case variable"
            )

    def regex_match_alphanumeric(self, str_, regex_match="^[a-zA-Z0-9_]+$"):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return re.match(regex_match, str_)

    def bl_has_numbers(self, str_):
        """
        DOCSTRING: CHECK WHETER THE STRING HAS NUMBERS
        INPUTS:
        OUTPUTS: BOOLEAN
        """
        return bool(re.search(r"\d", str_))

    def nfc_equal(self, str1, str2):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: UNICODE EQUIVALENCE TO IDENTIFY ENCODING STARDARDS THAT REPRESENT ESSENTIALLY
            THE SAME CHARACTER
        INPUTS: STRING 1 AND 2
        OUTPUTS: BOOLEAN
        """
        return normalize("NFC", str1) == normalize("NFC", str2)

    def casefold_equal(self, str1, str2):
        """
        REFERENCES: “FLUENT PYTHON BY LUCIANO RAMALHO (O’REILLY). COPYRIGHT 2015 LUCIANO RAMALHO, 978-1-491-94600-8.”
        DOCSTRING: UNICODE EQUIVALENCE TO IDENTIFY ENCODING STARDARDS THAT REPRESENT ESSENTIALLY
            THE SAME CASEFOLD FOR A GIVEN CHARACTER
        INPUTS: STRING 1 AND 2
        OUTPUTS: BOOLEAN
        """
        return normalize("NFC", str1).casefold() == normalize("NFC", str2).casefold()

    def remove_non_alphanumeric_chars(
        self, str_, str_pattern_maintain=r'[\W_]', str_replace=''
    ):
        """
        Remove non-alphanumeric characters from a string.
        Args:
            str_ (str): The input string.
            str_pattern_maintain (str): Regex pattern to match non-alphanumeric characters (default: r'[\W_]').
            str_replace (str): The string to replace matched characters with (default: '').
        Returns:
            str: The string with non-alphanumeric characters removed.
        """
        return re.sub(str_pattern_maintain, str_replace, str_)

    def remove_numeric_chars(self, str_):
        """
        REFERENCES: https://stackoverflow.com/questions/12851791/removing-numbers-from-str_
        DOCSTRING: REMOVE NUMERIC CHARACTERS
        INPUTS: STRING
        OUTPUTS: STRING
        """
        def_remove_digits = str.maketrans("", "", digits)
        return str_.translate(def_remove_digits)

    def is_capitalized(self, str_, bl_simple_validation=True):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS
        """
        # removing diacritcs
        str_ = self.remove_diacritics(str_)
        # removing non-alfanumeric characters
        str_ = self.remove_non_alphanumeric_chars(str_)
        #   returning wheter is capitalized or not
        try:
            if bl_simple_validation == True:
                if (str_[0].isupper() == True) and (str_[1].islower() == True):
                    return True
                else:
                    return False
            else:
                if (str_[0].isupper() == True) and (
                    all([l.islower() for l in str_[1:]])
                ):
                    return True
                else:
                    return False
        except:
            return False

    def split_re(self, str_, re_split=r"[;,\s]\s*"):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return re.split(re_split, str_)

    def replace_case_insensitive(self, str_, str_replaced, str_replace):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return re.sub(str_replaced, str_replace, str_, flags=re.IGNORECASE)

    def matchcase(self, str_):
        """
        REFERENCES: PYTHON COOKBOOK - DAVID BEASZLEY, BRIAN K. JONES
        DOCSTRING: MATCHASE SENSE
        INPUTS: WORD
        OUTPUTS: STRING
        """
        def replace(m):
            str_ = m.group()
            if str_.isupper():
                return str_.upper()
            elif str_.islower():
                return str_.lower()
            elif str_[0].isupper():
                return str_.capitalize()
            else:
                return str_
        return replace

    def replace_respecting_case(self, str_, str_replaced, str_replace):
        """
        REFERENCES: PYTHON COOKBOOK - DAVID BEASZLEY, BRIAN K. JONES
        DOCSTRING: MATCHASE SENSE
        INPUTS: WORD
        OUTPUTS: STRING
        """
        return re.sub(
            str_replaced, self.matchcase(str_replace), str_, flags=re.IGNORECASE
        )

    def replace_all(self, str_, dict_replacers):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        for i, j in dict_replacers.items():
            str_ = str_.replace(i, j)
        return str_

    def html_to_txt(self, html_):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        soup = BeautifulSoup(html_, features="lxml")
        return soup(html_)

    def extract_urls(self, str_):
        """
        DOCSTRING: LIST OF URLS IN A GIVEN STRING
        INPUTS: STRING
        OUTPUTS: LIST
        """
        # define a regular expression pattern to match URLs
        url_pattern = re.compile(r"https?://\S+|www\.\S+")
        # find all matches in the given str_
        list_urls = re.findall(url_pattern, str_)
        # return urls list
        return list_urls

    def is_word(self, _value):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS
        """
        try:
            float(_value)
            return False
        except ValueError:
            return True

    def convert_case(self, str_: str, from_case: str, to_case: str) -> str:
        """
        Converts a string between different naming conventions:
            - camelCase - 'camel'
            - PascalCase - 'pascal'
            - kebab-case - 'kebab'
            - UPPER_CONSTANT - 'upper_constant'
            - lower_constant - 'lower_constant'
            - UpperFirst - 'upper_first'
            - Default (words separated by spaces, hyphens or underscores) - 'default'

        Args:
            from_case (str): Current case of the string
            to_case (str): Desired case of the string

        Returns:
            str: Transformed string
        """
        # from case
        if from_case == "camel":
            words = re.sub(r"([a-z])([A-Z])", r"\1_\2", str_)
            words = re.sub(r"([a-zA-Z])(\d)", r"\1_\2", words)
            words = re.sub(r"(\d)([a-zA-Z])", r"\1_\2", words)
            words = words.lower().split("_")
        elif from_case == "pascal":
            words = re.sub(r"([a-z])([A-Z])", r"\1_\2", str_).lower().split("_")
        elif from_case == "kebab":
            words = str_.lower().split("-")
        elif from_case == "upper_constant" or from_case == "lower_constant":
            if "-" in str_:
                words = str_.lower().split("-")
            else:
                words = str_.lower().split("_")
        elif from_case == "upper_first":
            words = [str_[0].upper() + str_[1:].lower()]
        elif from_case == "default":
            str_ = str_.replace(" - ", " ")
            str_ = str_.replace("-", " ")
            str_ = str_.replace("_", " ")
            str_ = str_.replace("+", " ")
            str_ = str_.replace(" (", " ")
            str_ = str_.replace(") ", " ")
            str_ = str_.replace(r"\n", " ")
            words = str_.lower().split()
        else:
            raise ValueError(
                "Invalid from_case. Choose from ['camel', 'pascal', 'snake', 'kebab', 'upper_constant', 'lower_constant', 'upper_first']"
            )
        # converting to case
        if to_case == "camel":
            return words[0] + "".join(word.capitalize() for word in words[1:])
        elif to_case == "pascal":
            return "".join(word.capitalize() for word in words)
        elif to_case == "snake":
            return "_".join(words).lower()
        elif to_case == "kebab":
            return "-".join(words).lower()
        elif to_case == "upper_constant":
            return "_".join(words).upper()
        elif to_case == "lower_constant":
            return "_".join(words).lower()
        elif to_case == "upper_first":
            return words[0].capitalize()
        else:
            raise ValueError(
                "Invalid to_case. Choose from ['camel', 'pascal', 'snake', 'kebab', 'upper_constant', 'lower_constant', 'upper_first']"
            )

    def extract_info_between_braces(
        self, str_: str, str_pattern: str = r"\{\{(.*?)\}\}"
    ) -> str:
        return re.findall(str_pattern, str_)

    def fill_placeholders(self, str_: str, dict_placeholders: Dict[str, Any]) -> str:
        """
        Fill fstr named placeholders

        Args:
            str_ (str): fstr
            dict_placeholders (Dict[str, Any]): named placeholders

        Returns:
            str
        """
        list_placeholders = self.extract_info_between_braces(str_)
        for placeholder in list_placeholders:
            if placeholder in dict_placeholders:
                str_ = str_.replace(
                    f"{{{{{placeholder}}}}}", str(dict_placeholders[placeholder])
                )
            else:
                str_ = str_.replace(f"{{{{{placeholder}}}}}", f"{{{{{placeholder}}}}}")
        return str_

    def fill_zeros(self, str_prefix: str, int_num: int, total_length: int = 11) -> str:
        str_num = str(int_num)
        required_zeros = total_length - len(str_prefix) - len(str_num)
        if required_zeros < 0:
            raise ValueError("Total length is too small for the given inputs.")
        return f"{str_prefix}{'0' * required_zeros}{str_num}"

    def get_url_query(self, url: str, bl_include_fragment: bool = False) \
        -> Dict[str, Union[str, List[str]]]:
        """
        Extracts parameters from a URL's query string or fragment.

        Args:
            url (str): The URL to parse.
            bl_include_fragment (bool): Whether to include parameters from the fragment (after #).
            Defaults to False.

        Returns:
            Dict[str, Union[str, List[str]]]: A dictionary of parameters. Single-value parameters
            are returned as strings, while multi-value parameters are returned as lists of strings.
        """
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)
        if bl_include_fragment == True:
            fragment_params = parse_qs(parsed_url.fragment)
            query_params.update(fragment_params)
        return {
            key: value[0] if len(value) == 1 else value
            for key, value in query_params.items()
        }

    def has_no_letters(self, str_: str) -> bool:
        """Check if a string has no letters (A-Z, a-z)."""
        return not any(char.isalpha() for char in str_)



import platform
import subprocess
import psutil
from typing import Dict, List, Optional
import time

class DriveClassifier:
    """
    A utility class for classifying and retrieving information about disk drives.
    """

    @property
    def classify_drives(self) -> Dict[str, List[str]]:
        """
        Classifies hard drive units as 'local' or 'detachable'.

        Returns:
            Dict[str, List[str]]: A dictionary with two keys:
                - 'local': List of local drive paths.
                - 'detachable': List of detachable drive paths.
        """
        dict_drives = {
            "local": [],
            "detachable": []
        }
        psutil_partitions = psutil.disk_partitions()
        for partition in psutil_partitions:
            if 'removable' in partition.opts:
                dict_drives["detachable"].append(partition.device)
            else:
                dict_drives["local"].append(partition.device)
        return dict_drives

    def get_drive_usage(self, drive_path: str) -> Optional[Dict[str, int]]:
        """
        Retrieves usage statistics for a specific drive.

        Args:
            drive_path[str]: The path of the drive (e.g., 'C:\\').

        Returns:
            dict: A dictionary with keys 'total', 'used', and 'free' (in bytes).
                  Returns None if the drive is not found.
        """
        try:
            usage = psutil.disk_usage(drive_path)
            return {
                "total": usage.total,
                "used": usage.used,
                "free": usage.free
            }
        except FileNotFoundError:
            return None

    def get_drive_info(self) -> List[Dict[str, str]]:
        """
        Retrieves detailed information about all drives.

        Returns:
            list: A list of dictionaries, each containing drive information such as
                  device, mount point, file system type, and options.
        """
        psutil_partitions = psutil.disk_partitions()
        drive_info = []
        for partition in psutil_partitions:
            drive_info.append({
                "device": partition.device,
                "mountpoint": partition.mountpoint,
                "fstype": partition.fstype,
                "opts": partition.opts
            })
        return drive_info

    def list_all_drives(self) -> List[str]:
        """
        Lists all drives without classification.

        Returns:
            list: A list of all drive paths.
        """
        psutil_partitions = psutil.disk_partitions()
        return [partition.device for partition in psutil_partitions]

    def filter_drives_by_fs(self, fs_type: str) -> List[str]:
        """
        Filters drives by their file system type.

        Args:
            fs_type[str]: The file system type to filter by (e.g., 'NTFS', 'ext4').

        Returns:
            list: A list of drive paths that match the specified file system type.
        """
        psutil_partitions = psutil.disk_partitions()
        return [partition.device for partition in psutil_partitions if partition.fstype == fs_type]

    def monitor_drive_changes(self, interval: int = 5) -> None:
        """
        Monitors for changes in connected drives (e.g., USB drives being connected or disconnected).

        Args:
            interval (int): The time interval (in seconds) between checks.
        """
        previous_drives = set(self.list_all_drives())
        print("Monitoring drive changes. Press Ctrl+C to stop.")
        try:
            while True:
                time.sleep(interval)
                current_drives = set(self.list_all_drives())
                if current_drives != previous_drives:
                    added = current_drives - previous_drives
                    removed = previous_drives - current_drives
                    if added:
                        print(f"Drives added: {added}")
                    if removed:
                        print(f"Drives removed: {removed}")
                    previous_drives = current_drives
        except KeyboardInterrupt:
            print("Drive monitoring stopped.")

    def get_drive_serial(self, drive_path: str) -> Optional[str]:
        """
        Retrieves the serial number or unique identifier for a drive (platform-dependent).

        Args:
            drive_path (str): The path of the drive (e.g., 'C:\\' on Windows, '/dev/sda' on Linux).

        Returns:
            str: The serial number or unique identifier of the drive.
                 Returns None if the serial number cannot be retrieved.
        """
        system = platform.system()
        if system == "Windows":
            return self._get_drive_serial_windows(drive_path)
        elif system == "Linux":
            return self._get_drive_serial_linux(drive_path)
        # macOS
        elif system == "Darwin":
            return self._get_drive_serial_mac(drive_path)
        else:
            print(f"Unsupported operating system: {system}")
            return None

    def _get_drive_serial_windows(self, drive_path: str) -> Optional[str]:
        """
        Retrieves the serial number for a drive on Windows.

        Args:
            drive_path (str): The path of the drive (e.g., 'C:\\').

        Returns:
            str: The serial number of the drive.
        """
        try:
            import wmi
            c = wmi.WMI()
            for disk in c.Win32_DiskDrive():
                if disk.DeviceID == drive_path:
                    return disk.SerialNumber.strip()
            return None
        except ImportError:
            print("The 'wmi' library is required for this functionality on Windows.")
            return None

    def _get_drive_serial_linux(self, drive_path: str) -> Optional[str]:
        """
        Retrieves the serial number for a drive on Linux.

        Args:
            drive_path (str): The path of the drive (e.g., '/dev/sda').

        Returns:
            str: The serial number of the drive.
        """
        try:
            # Use `hdparm` or `lsblk` to get the serial number
            result = subprocess.run(
                ["sudo", "hdparm", "-I", drive_path],
                capture_output=True,
                text=True
            )
            if result.returncode == 0:
                for line in result.stdout.splitlines():
                    if "Serial Number:" in line:
                        return line.split(":")[1].strip()
            return None
        except FileNotFoundError:
            print("The 'hdparm' command is required for this functionality on Linux.")
            return None

    def _get_drive_serial_mac(self, drive_path: str) -> Optional[str]:
        """
        Retrieves the serial number for a drive on macOS.

        Args:
            drive_path (str): The path of the drive (e.g., '/dev/disk0').

        Returns:
            str: The serial number of the drive.
        """
        try:
            # Use `diskutil` to get the serial number
            result = subprocess.run(
                ["diskutil", "info", drive_path],
                capture_output=True,
                text=True
            )
            if result.returncode == 0:
                for line in result.stdout.splitlines():
                    if "Device / Media Name:" in line:
                        return line.split(":")[1].strip()
            return None
        except FileNotFoundError:
            print("The 'diskutil' command is required for this functionality on macOS.")
            return None


import pandas as pd
from requests import Response
from sqlalchemy.orm import Session
from datetime import datetime
from logging import Logger
from time import sleep
from typing import List, Optional
from stpstone._config.global_slots import YAML_BR_CVM_REGISTRIES
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy


class CVMRegistries(ABCRequests):

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None,
    ) -> None:
        super().__init__(
            dict_metadata=YAML_BR_CVM_REGISTRIES,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs,
        )
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs
        self.month_ref = self.dt_ref.strftime("%Y%m")
        self.month_ref = DatesBR().add_months(self.dt_ref, -1).strftime("%Y%m")
        self.year_ref = DatesBR().year_number(self.dt_ref)

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        return None



import re
import pandas as pd
from datetime import datetime
from typing import Optional, List
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from time import sleep
from stpstone._config.global_slots import YAML_WW_RATINGS_CORP_S_AND_P
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.ingestion.abc.requests import ABCRequests
from stpstone.utils.webdriver_tools.selenium_wd import SeleniumWD
from stpstone.utils.loggs.create_logs import CreateLog


class RatingsCorpSPGlobalConcreteCreator(ABCRequests):

    def __init__(
        self,
        bearer: str,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None,
        pg_number: int = 1
    ) -> None:
        super().__init__(
            dict_metadata=YAML_WW_RATINGS_CORP_S_AND_P,
            session=session,
            dt_ref=dt_ref,
            cls_db=cls_db,
            logger=logger,
            token=token,
            list_slugs=list_slugs
        )
        self.bearer = bearer
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.list_slugs = list_slugs
        self.pg_number = pg_number

    @property
    def get_bearer(self) -> str:
        regex_pattern = "(?i)system_access_token=([^*]+?);"
        url = "https://disclosure.spglobal.com/ratings/en/regulatory/ratings-actions"
        cls_selenium = SeleniumWD(url, bl_headless=True, bl_incognito=True)
        cls_selenium.wait(60)
        cls_selenium.wait_until_el_loaded('//a[@class="link-black link-black-hover text-underline"]')
        sleep(60)
        list_network_traffic = cls_selenium.get_network_traffic
        for i, dict_ in enumerate(list_network_traffic):
            if dict_["method"] == "Network.responseReceivedExtraInfo":
                int_idx_bearer = i
                break
        regex_match = re.search(
            regex_pattern,
            list_network_traffic[int_idx_bearer]["params"]["headers"]["set-cookie"]
        )
        if (regex_match is not None) \
            and (regex_match.group(0) is not None) \
            and (len(regex_match.group(0)) > 0):
            return f"Bearer {regex_match.group(1)}"

    def req_trt_injection(self, resp_req: Response) -> Optional[pd.DataFrame]:
        return pd.DataFrame(resp_req.json()["RatingAction"])


class RatingsCorpSPGlobalProduct:

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_ref: datetime = DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db: Optional[Session] = None,
        logger: Optional[Logger] = None,
        token: Optional[str] = None,
        list_slugs: Optional[List[str]] = None,
        pg_number: int = 1
    ) -> None:
        self.session = session
        self.dt_ref = dt_ref
        self.cls_db = cls_db
        self.logger = logger
        self.token = token
        self.list_slugs = list_slugs
        self.pg_number = pg_number

    @property
    def get_corp_ratings(self) -> pd.DataFrame:
        list_ser = list()
        str_bearer = RatingsCorpSPGlobalConcreteCreator(bearer=None).get_bearer
        for i in range(1, 100):
            try:
                cls_ = RatingsCorpSPGlobalConcreteCreator(pg_number=i, bearer=str_bearer)
                df_ = cls_.source("ratings_corp", bl_fetch=True)
                list_ser.extend(df_.to_dict(orient="records"))
                sleep(10)
            except Exception as e:
                CreateLog().log_message(self.logger, f"Error: {e}", log_level="warning")
        return pd.DataFrame(list_ser)

    @property
    def update_db(self) -> None:
        str_bearer = RatingsCorpSPGlobalConcreteCreator(bearer=None).get_bearer
        for i in range(1, 100):
            try:
                cls_ = RatingsCorpSPGlobalConcreteCreator(pg_number=i, bearer=str_bearer)
                _ = cls_.source("ratings_corp", bl_fetch=False)
                sleep(10)
            except Exception as e:
                CreateLog().log_message(self.logger, f"Error: {e}", log_level="warning")



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class Converter(metaclass=TypeChecker):

    def binary_to_decimal(self, binary: str) -> int:
        """
        Converts a binary string to a decimal integer.

        Args:
            binary: A string of binary digits (0s and 1s)

        Returns:
            int: The decimal integer equivalent of the binary string
        """
        return int(binary, 2)

    def decimal_to_binary(self, decimal: int) -> str:
        """
        Converts a decimal integer to a binary string. Remove the '0b' prefix

        Args:
            decimal: An integer

        Returns:
            str: The binary string equivalent of the decimal integer
        """
        return bin(decimal)[2:]

    def decimal_to_hexadecimal(self, decimal: int) -> str:
        """
        Converts a decimal integer to a hexadecimal string. Remove the '0x' prefix and convert
        to uppercase

        Args:
            decimal: An integer

        Returns:
            str: The hexadecimal string equivalent of the decimal integer, in uppercase
        """
        return hex(decimal)[2:].upper()

    def hexadecimal_to_decimal(self, hexadecimal: str) -> int:
        """
        Converts a hexadecimal string to a decimal integer.

        Args:
            hexadecimal: A string of hexadecimal digits (0-9, A-F)

        Returns:
            int: The decimal integer equivalent of the hexadecimal string
        """
        return int(hexadecimal, 16)


if __name__ == "__main__":
    converter = Converter()
    print(f"Binary to Decimal (1010) => {converter.binary_to_decimal('1010')}")  # output: 10
    print(f"Decimal to Binary (10) => {converter.decimal_to_binary(10)}")  # output: 1010
    print(f"Decimal to Hexadecimal (255) => {converter.decimal_to_hexadecimal(255)}")  # output: FF
    print(f"Hexadecimal to Decimal (FF) => {converter.hexadecimal_to_decimal('FF')}")  # output: 255
